from typing import List
import h5py
from ewokscore import Task

from .utils.data_utils import save_image


class SaveImages(
    Task,
    input_names=["image_uris", "output_dir"],
    optional_input_names=["ext"],
    output_names=["output_paths"],
):
    """Save images with monitor and other metadata"""

    def run(self):
        image_uris: List[str] = self.inputs.image_uris
        output_dir: str = self.inputs.output_dir
        ext: str = self.get_input_value("ext", "edf")

        output_paths = []

        for image_uri in image_uris:
            image_filename, image_h5path = image_uri.split("::")

            with h5py.File(image_filename, "r") as h5file:
                image_dset = h5file[image_h5path]

                output_paths.append(
                    save_image(
                        image_dset[()],
                        output_dir,
                        save_name=image_h5path.split("/")[-1],
                        monitor_data=image_dset.attrs.get("monitor", None),
                        metadata={
                            k: v for k, v in image_dset.attrs.items() if k != "monitor"
                        },
                        ext=ext,
                    )
                )

        self.outputs.output_paths = output_paths



import numpy
import h5py
import pytest
from ewoksorange.tests.utils import execute_task

from orangecontrib.ewoksxrpd.batchintegrate import OWIntegrateBlissScan


@pytest.mark.parametrize("ndims", [1, 2])
@pytest.mark.parametrize("external", [False, True])
def test_batch_integrate_task(ndims, external, bliss_task_inputs, tmpdir):
    assert_batch_integrate(ndims, bliss_task_inputs, tmpdir, external=external)


@pytest.mark.parametrize("ndims", [1, 2])
@pytest.mark.parametrize("external", [False, True])
def test_batch_integrate_widget(ndims, external, bliss_task_inputs, tmpdir, qtapp):
    assert_batch_integrate(
        ndims, bliss_task_inputs, tmpdir, external=external, qtapp=qtapp
    )


def assert_batch_integrate(
    ndims, bliss_task_inputs, tmpdir, external: bool = False, qtapp=None
):
    inputs = dict(bliss_task_inputs)
    output_filename = str(tmpdir / "result.h5")
    inputs["output_filename"] = output_filename
    if external:
        external_output_filename = str(tmpdir / "data.h5")
        inputs["external_output_filename"] = external_output_filename
    if ndims == 2:
        inputs["integration_options"] = {
            "method": "no_csr_cython",
            "integrator_name": "integrate2d_ng",
            "nbpt_azim": 100,
            "error_model": "poisson",
        }
    else:
        inputs["integration_options"] = {
            "error_model": "azimuthal",
            "method": "no_csr_cython",
            "integrator_name": "sigma_clip_ng",
            "extra_options": {"max_iter": 3, "thres": 0},
        }

    outputs = execute_task(
        OWIntegrateBlissScan.ewokstaskclass if qtapp is None else OWIntegrateBlissScan,
        inputs=inputs,
    )

    if external:
        expected = {
            "nxdata_url": f"{external_output_filename}::/2.1/p3_integrate/integrated"
        }
    else:
        expected = {"nxdata_url": f"{output_filename}::/2.1/p3_integrate/integrated"}
    assert outputs == expected

    with h5py.File(output_filename) as root:
        data = root["2.1/measurement/p3_integrated"][()]
        if ndims == 2:
            axes = ["points", "chi", "2th"]
        else:
            axes = ["points", "2th"]
        assert root["2.1/p3_integrate/integrated"].attrs["axes"].tolist() == axes
        spectrum0 = data[0]
        for spectrum in data:
            numpy.testing.assert_allclose(spectrum, spectrum0, atol=1)

        # Check links to raw data
        root["2.1/measurement/monitor"][()]
        list(root["2.1/instrument"].keys())



import numpy
import pyFAI.azimuthalIntegrator
from ewoksorange.tests.utils import execute_task

from .xrpd_theory import Measurement

from orangecontrib.ewoksxrpd.mask import OWMaskDetection


def test_mask_detection_task(
    image1Setup1SampleB: Measurement,
    image2Setup1SampleB: Measurement,
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
):
    assert_mask_detection(image1Setup1SampleB, image2Setup1SampleB, aiSetup1, None)


def test_mask_detection_widget(
    image1Setup1SampleB: Measurement,
    image2Setup1SampleB: Measurement,
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    qtapp,
):
    assert_mask_detection(image1Setup1SampleB, image2Setup1SampleB, aiSetup1, qtapp)


def assert_mask_detection(
    image1Setup1SampleB: Measurement,
    image2Setup1SampleB: Measurement,
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    qtapp,
):
    inputs = {
        "image1": image1Setup1SampleB.image,
        "monitor1": image1Setup1SampleB.monitor,
        "image2": image2Setup1SampleB.image,
        "monitor2": image2Setup1SampleB.monitor,
    }
    results = execute_task(
        OWMaskDetection.ewokstaskclass if qtapp is None else OWMaskDetection,
        inputs=inputs,
    )

    expected = aiSetup1.detector.get_mask()
    numpy.testing.assert_array_equal(results["mask"], expected)

    inputs["smooth"] = 5
    results = execute_task(
        OWMaskDetection.ewokstaskclass if qtapp is None else OWMaskDetection,
        inputs=inputs,
    )

    borders = results["mask"] - expected
    assert not (borders == 0).all(), "smoothing should add masked pixels"
    assert not (borders < 0).any(), "smoothing should never remove masked pixels"



import os
import h5py
import pytest
from contextlib import ExitStack

from ..tasks.utils import data_utils


@pytest.mark.parametrize("relative", [True, False], ids=["relative", "absolute"])
def test_hdf5_external_link(tmp_path, relative):
    topdir = tmp_path / "rootdir"
    middir = topdir / "middir"
    lowdir = middir / "lowdir"
    lowdir.mkdir(parents=True)

    with ExitStack() as stack:
        external_up = stack.enter_context(
            h5py.File(str(topdir / "external_up.h5"), mode="a")
        )
        master = stack.enter_context(h5py.File(str(middir / "master.h5"), mode="a"))
        external_down = stack.enter_context(
            h5py.File(str(lowdir / "external_down.h5"), mode="a")
        )

        external_up["external_data"] = 1
        master["/1.1/instrument/diode2/data"] = 2
        external_down["external_data"] = 3

        # Down-link to external data
        link = data_utils.hdf5_link_object(
            master.filename,
            "/1.1/instrument/diode3",
            "data",
            external_down.filename,
            "external_data",
            relative=relative,
        )
        assert type(link) is h5py.ExternalLink
        assert link.path == "external_data"
        if relative:
            assert link.filename == os.path.join("lowdir", "external_down.h5")
        else:
            assert link.filename == external_down.filename

        data_utils.create_hdf5_link(
            master,
            "/1.1/instrument/diode3/data",
            external_down["external_data"],
            relative=relative,
            raise_on_exists=True,
        )

        # Up-link to external data
        link = data_utils.hdf5_link_object(
            master.filename,
            "/1.1/instrument/diode1",
            "data",
            external_up.filename,
            "external_data",
            relative=relative,
        )
        assert type(link) is h5py.ExternalLink
        assert link.path == "external_data"
        if relative:
            assert link.filename == os.path.join("..", "external_up.h5")
        else:
            assert link.filename == external_up.filename

        data_utils.create_hdf5_link(
            master,
            "/1.1/instrument/diode1/data",
            external_up["external_data"],
            relative=relative,
            raise_on_exists=True,
        )

        # Create external link again
        data_utils.create_hdf5_link(
            master,
            "/1.1/instrument/diode3/data",
            external_down["external_data"],
            relative=relative,
        )
        with pytest.raises(Exception, match="name already exists"):
            data_utils.create_hdf5_link(
                master,
                "/1.1/instrument/diode3/data",
                external_down["external_data"],
                relative=relative,
                raise_on_exists=True,
            )

        # Check whether links resolve
        assert master["/1.1/instrument/diode1/data"][()] == 1
        assert master["/1.1/instrument/diode2/data"][()] == 2
        assert master["/1.1/instrument/diode3/data"][()] == 3

    # Check whether links after moving the files
    topdir.rename(tmp_path / "newname")
    topdir = tmp_path / "newname"
    middir = topdir / "middir"
    lowdir = middir / "lowdir"

    with h5py.File(str(middir / "master.h5"), mode="r") as master:
        assert master["/1.1/instrument/diode2/data"][()] == 2
        if relative:
            assert master["/1.1/instrument/diode1/data"][()] == 1
            assert master["/1.1/instrument/diode3/data"][()] == 3
        else:
            with pytest.raises(KeyError):
                master["/1.1/instrument/diode1/data"]
            with pytest.raises(KeyError):
                master["/1.1/instrument/diode3/data"]


@pytest.mark.parametrize("relative", [True, False], ids=["relative", "absolute"])
def test_hdf5_internal_link(tmp_path, relative):

    with h5py.File(str(tmp_path / "master.h5"), mode="a") as root:
        root["data"] = 1
        root["midgroup/data2"] = 2
        root["midgroup/lowgroup/data"] = 3
        internal_up = root
        master = root["midgroup"]
        internal_down = root["midgroup/lowgroup"]

        # Link to self
        link = data_utils.hdf5_link_object(
            root.filename, "", "data", root.filename, "/data", relative=relative
        )
        assert link is None
        link = data_utils.hdf5_link_object(
            root.filename, "/", "data", root.filename, "/data", relative=relative
        )
        assert link is None
        link = data_utils.hdf5_link_object(
            root.filename, "/", "data", root.filename, "data", relative=relative
        )
        assert link is None
        link = data_utils.hdf5_link_object(
            root.filename, "", "data", root.filename, "data", relative=relative
        )
        assert link is None
        data_utils.create_hdf5_link(
            root, "/data", root["data"], relative=relative, raise_on_exists=True
        )

        # Down-link to internal data
        link = data_utils.hdf5_link_object(
            root.filename,
            "/midgroup",
            "data3",
            root.filename,
            "/midgroup/lowgroup/data",
            relative=relative,
        )
        assert type(link) is h5py.SoftLink
        if relative:
            assert link.path == "lowgroup/data"
        else:
            assert link.path == "/midgroup/lowgroup/data"

        data_utils.create_hdf5_link(
            master,
            "/midgroup/data3",
            internal_down["data"],
            relative=relative,
            raise_on_exists=True,
        )

        # Up-link to internal data
        link = data_utils.hdf5_link_object(
            root.filename,
            "/midgroup/data1",
            "data",
            root.filename,
            "data",
            relative=relative,
        )
        assert type(link) is h5py.SoftLink
        assert link.path == "/data"

        data_utils.create_hdf5_link(
            master,
            "/midgroup/data1",
            internal_up["data"],
            relative=relative,
            raise_on_exists=True,
        )

        # Create internal link again
        data_utils.create_hdf5_link(
            master, "/midgroup/data3", internal_down["data"], relative=relative
        )
        with pytest.raises(Exception, match="name already exists"):
            data_utils.create_hdf5_link(
                master,
                "/midgroup/data3",
                internal_down["data"],
                relative=relative,
                raise_on_exists=True,
            )

        # Check whether links resolve
        assert master["/midgroup/data1"][()] == 1
        assert master["/midgroup/data2"][()] == 2
        assert master["/midgroup/data3"][()] == 3



import logging
from ewoksxrpd.tasks.log import zip_with_progress


def test_zip_with_progress(caplog):
    with caplog.at_level(logging.INFO):
        values = list()
        for value in zip_with_progress([], message="Integrated %s images of %s"):
            values.append(value)

    assert not values
    assert "Integrated 0 images of ? (FINISHED)" in caplog.text

    with caplog.at_level(logging.INFO):
        values = list()
        for (value,) in zip_with_progress([1], message="Integrated %s images of %s"):
            values.append(value)

    assert values == [1]
    assert "Integrated 1 images of ? (FINISHED)" in caplog.text

    with caplog.at_level(logging.INFO):
        values = list()
        for value, *_ in zip_with_progress(
            [1, 2], message="Integrated %s images of %s"
        ):
            values.append(value)

    assert values == [1, 2]
    assert "Integrated 2 images of ? (FINISHED)" in caplog.text

    with caplog.at_level(logging.INFO):
        values = list()
        for nr, letter in zip_with_progress(
            [1, 2],
            [
                "a",
            ],
            message="Integrated %s images of %s",
        ):
            values.append((nr, letter))

    assert values == [(1, "a")]
    assert "Integrated 1 images of ? (FINISHED)" in caplog.text



import os
from typing import Mapping, Union, Optional
from typing_extensions import TypeGuard
from numbers import Number

import numpy
import h5py
from silx.io.url import DataUrl
from blissdata.h5api import dynamic_hdf5
import fabio.edfimage
import fabio.tifimage


def hdf5_url(file_name: str, data_path: str) -> str:
    if not os.path.isabs(file_name):
        file_name = os.path.abspath(file_name)
    return f"silx://{file_name}?path={data_path}"


def is_data(data) -> TypeGuard[Union[numpy.ndarray, Number, str, list]]:
    if isinstance(data, (numpy.ndarray, Number)):
        return True
    if isinstance(data, (str, list)) and data:
        return True
    return False


def data_from_storage(data, remove_numpy=True):
    if isinstance(data, numpy.ndarray):
        if not remove_numpy:
            return data
        elif data.ndim == 0:
            return data.item()
        else:
            return data.tolist()
    elif isinstance(data, Mapping):
        return {
            k: data_from_storage(v, remove_numpy=remove_numpy)
            for k, v in data.items()
            if not k.startswith("@")
        }
    else:
        return data


def create_hdf5_link(
    parent: h5py.Group,
    link_name: str,
    target: Union[h5py.Dataset, h5py.Group],
    relative: bool = True,
    raise_on_exists: bool = False,
) -> None:
    """
    :param parent: HDF5 group in which the link will be created
    :param link_name: relative HDF5 path of the link source with respect to :code:`parent`
    :param target: absolute HDF5 path of the link target
    :param relative: determines whether the external or internal link is absolute or relative.
                     Internal links that refer upwards are not supported and will always be absolute.
    :param raise_on_exists: raise exception when :code:`link_name` already exists
    """
    link = hdf5_link_object(
        _get_hdf5_filename(parent.file),
        parent.name,
        link_name,
        _get_hdf5_filename(target.file),
        target.name,
        relative=relative,
    )
    if link is None:
        # Link refers to itself
        return
    if link_name in parent and not raise_on_exists:
        # Name already exists and is not necessarily equivalent to the link we want to create
        return
    parent[link_name] = link


def _get_hdf5_filename(file_obj) -> str:
    try:
        return file_obj.filename
    except AttributeError:
        # to be fixed in blissdata
        return file_obj._retry_handler.file_obj.filename


def hdf5_link_object(
    parent_filename: str,
    parent_name: str,
    link_name: str,
    target_filename: str,
    target_name: str,
    relative: bool = True,
) -> Union[None, h5py.ExternalLink, h5py.SoftLink]:
    """
    :param parent_filename: HDF5 filename in which the link will be created
    :param parent_name: absolute HDF5 group path in :code:`parent_filename`
    :param link_name: relative HDF5 path of the link source with respect to :code:`parent_name`
    :param target_filename: HDF5 filename in which the link target is located
    :param target_name: absolute HDF5 path in :code:`target_filename` of the link target
    :param relative: determines whether the external or internal link is absolute or relative.
                     Internal links that refer upwards are not supported and will always be absolute.
    :returns: Internal or external link object to be used to create the HDF5 link.
              Returns :code:`None` when the link refers to itself.
    """
    abs_parent_filename = os.path.abspath(parent_filename)
    target_is_absolute = os.path.isabs(target_filename)
    if target_is_absolute:
        abs_target_filename = target_filename
    else:
        abs_target_filename = os.path.join(
            os.path.dirname(abs_parent_filename), target_filename
        )

    # Internal link
    if os.path.normpath(abs_parent_filename) == os.path.normpath(abs_target_filename):
        link_full_name = _normalize_hdf5_item_name(parent_name, link_name)
        target_name = _normalize_hdf5_item_name(target_name)
        if link_full_name == target_name:
            # Link refers to itself
            return
        if not relative:
            return h5py.SoftLink(target_name)
        rel_target_name = os.path.relpath(target_name, parent_name)
        if ".." in rel_target_name:
            # Internal links upwards are not supported
            return h5py.SoftLink(target_name)
        return h5py.SoftLink(rel_target_name)

    # External link
    if relative:
        target_filename = os.path.relpath(
            abs_target_filename, os.path.dirname(abs_parent_filename)
        )
    else:
        target_filename = abs_target_filename
    return h5py.ExternalLink(target_filename, target_name)


def _normalize_hdf5_item_name(*parts) -> str:
    name = "/".join([s for part in parts for s in part.split("/") if s])
    return f"/{name}"


def link_bliss_scan(
    outentry: h5py.Group, bliss_scan_url: Union[str, DataUrl], **options
):
    if isinstance(bliss_scan_url, str):
        bliss_scan_url = DataUrl(bliss_scan_url)
    file_path = bliss_scan_url.file_path()
    data_path = bliss_scan_url.data_path()
    out_filename = outentry.file.filename
    ext_filename = os.path.relpath(out_filename, os.path.dirname(file_path))
    if ".." in ext_filename:
        ext_filename = file_path
    with dynamic_hdf5.File(file_path, mode="r", **options) as root:
        inentry = root[data_path]
        # Link to the entire group
        for groupname in ("instrument", "sample"):
            try:
                if groupname in outentry or groupname not in inentry:
                    continue
            except Exception:  # fixed by bliss PR !5435
                continue
            create_hdf5_link(outentry, groupname, inentry[groupname])

        # Link to all sub groups
        for groupname in ("measurement",):
            if groupname not in inentry:
                continue
            igroup = inentry[groupname]
            if groupname in outentry:
                ogroup = outentry[groupname]
            else:
                ogroup = outentry.create_group(groupname)
                ogroup.attrs["NX_class"] = igroup.attrs["NX_class"]
            for name in igroup.keys():
                if name in ogroup:
                    continue
                if name not in ogroup:
                    create_hdf5_link(ogroup, name, igroup[name])


def convert_to_3d(data: Union[numpy.ndarray, Number, str, list]):
    data_arr = numpy.array(data)

    if data_arr.ndim >= 3:
        return data_arr

    if data_arr.ndim == 2:
        return data_arr.reshape(1, *data_arr.shape)

    if data_arr.ndim == 1:
        return data_arr.reshape(1, 1, *data_arr.shape)

    return data_arr.reshape(1, 1, 1)


def save_image(
    data: numpy.ndarray,
    save_path: str,
    save_name: str,
    monitor_data: Optional[Number] = None,
    metadata: Optional[dict] = None,
    ext: str = "edf",
) -> str:
    normalized_data = data / monitor_data if monitor_data else data

    header = dict(
        Summed_monitor_counts=str(monitor_data) if monitor_data else "nan",
        **(metadata if metadata else {}),
    )

    ext = ext.lower()
    img_filepath = f"{save_path}/{save_name}.{ext}"
    if ext.lower() == "edf":
        Image = fabio.edfimage.EdfImage
    elif ext == "tiff" or ext == "tif":
        Image = fabio.tifimage.TifImage
    else:
        raise ValueError(f"Unsupported ext {ext}. Only supports EDF and TIFF.")

    img = Image(data=normalized_data, header=header)
    img.write(img_filepath)

    return img_filepath



import re
import json
from typing import Union, Dict, Optional, Any, Iterator, Tuple, List, NamedTuple
from collections.abc import Mapping, Sequence

import numpy
import h5py
from pyFAI import version as pyfai_version
from ewoks import __version__ as ewoks_version
from pyFAI.io.ponifile import PoniFile
from pyFAI.units import Unit
from silx.io.dictdump import dicttonx
from ewokscore.task import TaskInputError
from ewokscore.missing_data import is_missing_data

from .xrpd_utils import energy_wavelength
from .data_utils import data_from_storage
from .data_utils import create_hdf5_link


_REPLACE_PATTERNS = {
    "/gpfs/[^/]+/": "/",
    "/mnt/multipath-shares/": "/",
    "/lbsram/": "/",
}


class AxisInfo(NamedTuple):
    name: str
    units: str

    def to_str(self):
        return f"{self.name}_{self.units}"


def parse_pyfai_units(
    pyfai_unit: Union[Unit, Tuple[Unit, Unit]]
) -> Tuple[AxisInfo, AxisInfo]:
    """
    Parse PyFAI result units into a tuple containing the radial axis info and the azimuthal axis info.
    Handles pyFAI >= 2024.1 units (tuple with radial+azimuthal) and pyFAI < 2024.1 units (only radial unit).

    If no info is available for the azimuthal axis, it will default to name "chi" and unit "deg".
    """
    if isinstance(pyfai_unit, tuple):
        return _parse_pyfai_unit(pyfai_unit[0]), _parse_pyfai_unit(pyfai_unit[1])

    return _parse_pyfai_unit(pyfai_unit), AxisInfo("chi", "deg")


def _parse_pyfai_unit(pyfai_unit: Unit) -> AxisInfo:
    unit_tuple = tuple(pyfai_unit.name.split("_"))
    if len(unit_tuple) != 2:
        raise ValueError(f"Expected unit to be of the form X_Y. Got {pyfai_unit.name}")

    return AxisInfo(*unit_tuple)


def parse_string_units(unit: Union[str, numpy.ndarray]) -> AxisInfo:
    if isinstance(unit, numpy.ndarray):
        unit = unit.item()
    if not isinstance(unit, str):
        raise TypeError(type(unit))
    unit_tuple = tuple(unit.split("_"))
    if len(unit_tuple) != 2:
        raise ValueError(f"Expected unit to be of the form X_Y. Got {unit}")
    return AxisInfo(*unit_tuple)


def read_config(
    filename: Optional[str], replace_patterns: Optional[Dict[str, str]] = None
) -> dict:
    if not filename:
        return dict()
    if filename.endswith(".json"):
        parameters = _read_json(filename)
    else:
        parameters = _read_poni(filename)
    return normalize_parameters(parameters, replace_patterns=replace_patterns)


def _read_json(filename: str) -> dict:
    with open(filename, "r") as fp:
        return json.load(fp)


def _read_poni(filename: str) -> dict:
    return PoniFile(filename).as_dict()


def normalize_parameters(
    parameters: Union[str, int, float, Mapping, Sequence],
    replace_patterns: Optional[Dict[str, str]] = None,
) -> Union[str, int, float, Mapping, Sequence]:
    if replace_patterns is None:
        replace_patterns = _REPLACE_PATTERNS
    if isinstance(parameters, str):
        for pattern, repl in replace_patterns.items():
            parameters = re.sub(pattern, repl, parameters)
        return parameters
    if isinstance(parameters, Mapping):
        return {
            k: normalize_parameters(v, replace_patterns=replace_patterns)
            for k, v in parameters.items()
        }
    if isinstance(parameters, Sequence):
        return [
            normalize_parameters(v, replace_patterns=replace_patterns)
            for v in parameters
        ]
    return parameters


def extract_mask(integration_options: dict) -> Optional[numpy.ndarray]:
    """Integration options related to the mask:

    * mask_file: URL or None
    * do_mask: use mask_file or not
    * mask: URL, array or None (overwrites mask_file and do_mask when not None)
    """
    # Typically from a JSON configuration file
    mask_file = integration_options.get("mask_file", None)
    do_mask = integration_options.get("do_mask", None)
    if not do_mask:
        mask_file = None

    # Typically in memory from a previous calculation
    mask = integration_options.pop("mask", None)
    if isinstance(mask, str):
        mask_file = mask
        mask = None
        do_mask = True

    # Let pyFAI handle URLs (if any) and return in-memory (if any)
    integration_options["mask_file"] = mask_file
    integration_options["do_mask"] = do_mask
    return mask


def extract_flatfield(integration_options: dict) -> Optional[numpy.ndarray]:
    """Integration options related to the flatfield:

    * flat_field: URL(s) (list of strings or comma separate string) or None
    * do_flat: use flat_field or not
    * flatfield: URL(s) (list of strings or comma separate string), array or None (overwrites flat_field and do_flat when not None)

    When a numpy array is returned, the flat field is removed from the integration options.
    """
    # Typically from a JSON configuration file
    flat_field = integration_options.get("flat_field", None)
    do_flat = integration_options.get("do_flat", None)
    if not do_flat:
        flat_field = None
    if not flat_field:
        do_flat = False

    # Typically in memory from a previous calculation
    flatfield = integration_options.pop("flatfield", None)
    if isinstance(flatfield, str):
        flatfield = [s.strip() for s in flatfield.split(",")]
    if isinstance(flatfield, Sequence) and isinstance(flatfield[0], str):
        flat_field = flatfield
        flatfield = None
        do_flat = True
    if flatfield is not None:
        flat_field = None
        do_flat = False

    # Let pyFAI handle URLs (if any) and return in-memory (if any)
    integration_options["flat_field"] = flat_field
    integration_options["do_flat"] = do_flat
    return flatfield


def extract_darkcurrent(integration_options: dict) -> Optional[numpy.ndarray]:
    """Integration options related to the dark-current:

    * dark_current: URL(s) (list of strings or comma separate string) or None
    * do_dark: use flat_field or not
    * darkcurrent: URL(s) (list of strings or comma separate string), array or None (overwrites dark_current and do_dark when not None)

    When a numpy array is returned, the dark current is removed from the integration options.
    """
    # Typically from a JSON configuration file
    dark_current = integration_options.get("dark_current", None)
    do_dark = integration_options.get("do_dark", None)
    if not do_dark:
        dark_current = None
    if not dark_current:
        do_dark = False

    # Typically in memory from a previous calculation
    darkcurrent = integration_options.pop("darkcurrent", None)
    if isinstance(darkcurrent, str):
        darkcurrent = [s.strip() for s in darkcurrent.split(",")]
    if isinstance(darkcurrent, Sequence) and isinstance(darkcurrent[0], str):
        dark_current = darkcurrent
        darkcurrent = None
        do_dark = True
    if darkcurrent is not None:
        dark_current = None
        do_dark = False

    # Let pyFAI handle URLs (if any) and return in-memory (if any)
    integration_options["dark_current"] = dark_current
    integration_options["do_dark"] = do_dark
    return darkcurrent


def compile_integration_info(parameters: Mapping, **extra) -> Dict[str, Any]:
    """Compile information on a pyFAI integration process. Add and rename keys when appropriate."""
    integration_info = dict(parameters)

    mask = extract_mask(integration_info)
    flatfield = extract_flatfield(integration_info)
    darkcurrent = extract_darkcurrent(integration_info)

    # Do not save the in-memory make, flat field and dark current
    if mask is not None:
        integration_info["mask_file"] = "[...]"
    if flatfield is not None:
        integration_info["flat_field"] = "[...]"
    if darkcurrent is not None:
        integration_info["dark_current"] = "[...]"

    for k, v in extra.items():
        if v is not None:
            integration_info[k] = v
    wavelength = integration_info.get("wavelength")
    if wavelength is not None:
        integration_info["energy"] = energy_wavelength(wavelength)
    return integration_info


def integration_info_as_text(integration_info: Mapping, **extra) -> List[str]:
    """Convert to a flat list of strings with the format `{key} = {value}`.
    Add keys and units when appropriate.
    """
    flatdict = {"pyfai_version": pyfai_version, "ewoks_version": ewoks_version}
    flatdict.update(integration_info)
    _add_extra(flatdict, extra)
    flatdict = dict(_flatten_dict(flatdict))

    energy = flatdict.pop("energy", None)
    if energy:
        flatdict["energy"] = f"{energy:.18e} keV"

    wavelength = flatdict.pop("wavelength", None)
    if wavelength is not None:
        flatdict["wavelength"] = f"{wavelength:.18e} m"

    geometry_dist = flatdict.pop("geometry.dist", None)
    if geometry_dist is not None:
        flatdict["distance"] = f"{geometry_dist:.18e} m"

    geometry_poni1 = flatdict.pop("geometry.poni1", None)
    if geometry_poni1 is not None:
        flatdict["center dim0"] = f"{geometry_poni1:.18e} m"

    geometry_poni2 = flatdict.pop("geometry.poni2", None)
    if geometry_poni2 is not None:
        flatdict["center dim1"] = f"{geometry_poni2:.18e} m"

    geometry_rot1 = flatdict.pop("geometry.rot1", None)
    if geometry_rot1 is not None:
        flatdict["rot1"] = f"{geometry_rot1:.18e} rad"

    geometry_rot2 = flatdict.pop("geometry.rot2", None)
    if geometry_rot2 is not None:
        flatdict["rot2"] = f"{geometry_rot2:.18e} rad"

    geometry_rot3 = flatdict.pop("geometry.rot3", None)
    if geometry_rot3 is not None:
        flatdict["rot3"] = f"{geometry_rot3:.18e} rad"

    return [f"{k} = {v}" for k, v in flatdict.items()]


def integration_info_as_nxdict(
    integration_info, as_nxnote: bool = True
) -> Dict[str, Any]:
    """Convert to a Nexus dictionary. Add keys and units when appropriate."""
    configuration = {"ewoks_version": ewoks_version}
    configuration.update(integration_info)
    nxtree_dict = {
        "@NX_class": "NXprocess",
        "program": "pyFAI",
        "version": pyfai_version,
    }
    if as_nxnote:
        nxtree_dict["configuration"] = {
            "@NX_class": "NXnote",
            "type": "application/json",
            "data": json.dumps(configuration, cls=PyFaiEncoder),
        }
    else:
        configuration["@NX_class"] = "NXcollection"
        nxtree_dict["configuration"] = configuration
        if "energy" in configuration:
            configuration["energy@units"] = "keV"
        if "wavelength" in configuration:
            configuration["wavelength@units"] = "m"
        geometry = configuration.get("geometry", dict())
        if "dist" in geometry:
            geometry["dist@units"] = "m"
        if "poni1" in geometry:
            geometry["poni1@units"] = "m"
        if "poni2" in geometry:
            geometry["poni1@units"] = "m"
        if "rot1" in geometry:
            geometry["rot1@units"] = "rad"
        if "rot2" in geometry:
            geometry["rot2@units"] = "rad"
        if "rot3" in geometry:
            geometry["rot3@units"] = "rad"
    return nxtree_dict


class PyFaiEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (numpy.generic, numpy.ndarray)):
            return obj.tolist()
        return super().default(obj)


def _flatten_dict(
    adict: Mapping, _prefix: Optional[str] = None
) -> Iterator[Tuple[str, Any]]:
    if _prefix is None:
        _prefix = ""
    for k, v in adict.items():
        k = _prefix + k
        if isinstance(v, Mapping):
            yield from _flatten_dict(v, _prefix=f"{k}.")
        else:
            yield k, v


def _add_extra(adict: Mapping, extra: Mapping):
    for k, v in extra.items():
        if v is not None:
            adict[k] = v


def create_nxprocess(
    parent: h5py.Group, link_parent: h5py.Group, nxprocess_name: str, info
) -> h5py.Group:
    """Create NXprocess group in the external parent with link in the parent when these are different groups."""
    nxtree_dict = integration_info_as_nxdict(info)
    nxprocess_path = f"{parent.name}/{nxprocess_name}"
    dicttonx(nxtree_dict, parent.file, h5path=nxprocess_path, update_mode="modify")
    nxprocess = parent[nxprocess_name]
    create_hdf5_link(link_parent, nxprocess_name, nxprocess)
    return nxprocess


def create_integration_results_nxdata(
    nxprocess: h5py.Group,
    intensity_dim: int,
    radial,
    radial_units: str,
    azimuthal,
    azimuthal_units: str,
) -> h5py.Group:
    nxdata = nxprocess.create_group("integrated")
    nxdata.attrs["NX_class"] = "NXdata"
    nxprocess.attrs["default"] = "integrated"

    # Axes interpretation
    add_axes_to_nxdata(
        nxdata, intensity_dim, radial, radial_units, azimuthal, azimuthal_units
    )

    return nxdata


def add_axes_to_nxdata(
    nxdata: h5py.Group,
    intensity_dim: int,
    radial,
    radial_units,
    azimuthal,
    azimuthal_units,
):
    # Axes names and units
    radial_units = data_from_storage(radial_units, remove_numpy=True)
    try:
        radial_axis = parse_string_units(radial_units)
    except ValueError as e:
        raise TaskInputError(e)
    has_azimuth = not is_missing_data(azimuthal) and azimuthal is not None
    if has_azimuth:
        azimuthal_axis = parse_string_units(azimuthal_units)

    if has_azimuth and intensity_dim == 2:
        nxdata.attrs["axes"] = [azimuthal_axis.name, radial_axis.name]
        nxdata.attrs["interpretation"] = "image"
    elif has_azimuth and intensity_dim == 3:
        nxdata.attrs["axes"] = [".", azimuthal_axis.name, radial_axis.name]
        nxdata.attrs["interpretation"] = "image"
    elif not has_azimuth and intensity_dim == 2:
        nxdata.attrs["axes"] = [".", radial_axis.name]
        nxdata.attrs["interpretation"] = "spectrum"
    elif not has_azimuth and intensity_dim == 1:
        nxdata.attrs["axes"] = [radial_axis.name]
        nxdata.attrs["interpretation"] = "spectrum"
    else:
        raise ValueError("Unrecognized data")

    dset = nxdata.create_dataset(radial_axis.name, data=radial)
    dset.attrs["units"] = radial_axis.units
    if has_azimuth:
        dset = nxdata.create_dataset(azimuthal_axis.name, data=azimuthal)
        dset.attrs["units"] = azimuthal_axis.units



import numpy
import pytest

from ewoksxrpd.tasks.worker import persistent_worker


def test_worker_demo():
    integrate_options = {
        "method": "no_csr_cython",
        "nbpt_rad": 4096,
        "unit": "q_nm^-1",
        "error_model": "azimuthal",
        "dist": 0.05,
        "poni1": 0.1,
        "poni2": 0.1,
        "rot1": 0.17453292519943295,
        "rot2": 0,
        "rot3": 0,
        "detector": "Pilatus1M",
        "detector_config": None,
        "wavelength": 1.0332016536100021e-10,
        "integrator_name": "sigma_clip_ng",
        "extra_options": {"max_iter": 3, "thres": 0},
    }

    # Correct Pilatus1M shape
    image = numpy.zeros((1043, 981))

    # lima-camera-simulator < 1.9.10 does not support odd image
    # shapes so blissdemo adds a border
    demo_image = numpy.zeros((1044, 982))

    with persistent_worker(integrate_options, demo=False) as worker:
        worker.process(image)
        with pytest.raises(AssertionError):
            worker.process(demo_image)

    with persistent_worker(integrate_options, demo=True) as worker:
        worker.process(demo_image)
        with pytest.raises(AssertionError):
            worker.process(image)



from contextlib import ExitStack
from typing import Optional, Tuple, Union
import numpy
from pyFAI.containers import Integrate1dResult, Integrate2dResult
from ewoksdata.data.hdf5.dataset_writer import DatasetWriter
from . import pyfai_utils
import h5py


def get_yerror(result: Integrate1dResult) -> numpy.ndarray:
    if result.sigma is None:
        return numpy.full_like(result.intensity, numpy.nan)

    return result.sigma


def save_result(
    result: Union[Integrate1dResult, Integrate2dResult],
    intensity_writer: Optional[DatasetWriter],
    error_writer: Optional[DatasetWriter],
    nxprocess: h5py.Group,
    stack: ExitStack,
    flush_period: Optional[float] = None,
) -> Tuple[Optional[DatasetWriter], Optional[DatasetWriter], bool]:
    if intensity_writer is None:
        radial_axis, azimuthal_axis = pyfai_utils.parse_pyfai_units(result.unit)
        nxdata = pyfai_utils.create_integration_results_nxdata(
            nxprocess,
            result.intensity.ndim + 1,  # +1 for the scan dimension
            result.radial,
            radial_axis.to_str(),
            result.azimuthal if isinstance(result, Integrate2dResult) else None,
            azimuthal_axis.to_str() if isinstance(result, Integrate2dResult) else None,
        )
        nxdata.attrs["signal"] = "intensity"
        intensity_writer = stack.enter_context(
            DatasetWriter(nxdata, "intensity", flush_period=flush_period)
        )
        if result.sigma is not None:
            error_writer = stack.enter_context(
                DatasetWriter(nxdata, "intensity_errors", flush_period=flush_period)
            )

    flush = intensity_writer.add_point(result.intensity)
    if result.sigma is not None:
        flush |= error_writer.add_point(result.sigma)

    return intensity_writer, error_writer, flush



from typing import Dict, Mapping, Tuple

from ewoksxrpd.tasks.integrate import IntegrateBlissScan
from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.gui.forms import input_parameters_integrateblissscan
from ewoksxrpd.gui.forms import pack_geometry
from ewoksxrpd.gui.forms import unpack_geometry
from ewoksxrpd.gui.forms import unpack_enabled_geometry


__all__ = ["OWIntegrateBlissScan"]


class OWIntegrateBlissScan(OWTriggerWidget, ewokstaskclass=IntegrateBlissScan):
    name = "IntegrateBlissScan"
    description = "1D or 2D integrate data from one detector of a single Bliss scan"
    icon = "icons/widget.png"
    want_main_area = True

    def _init_forms(self):
        parameter_info = input_parameters_integrateblissscan(
            self.get_default_input_values(), saving=True
        )
        self._create_input_form(parameter_info)

    def _values_from_form(
        self, values: Mapping, checked: Dict[str, bool], output: bool = False
    ) -> Mapping:
        return pack_geometry(values, checked)

    def _values_to_form(
        self, values: Mapping, output: bool = False
    ) -> Tuple[Mapping, Dict[str, bool]]:
        return unpack_geometry(values)

    def _enabled_to_form(
        self, enabled: Dict[str, bool], output: bool = False
    ) -> Dict[str, bool]:
        return unpack_enabled_geometry(enabled)



import numpy
import matplotlib
import matplotlib.scale as mscale
import matplotlib.transforms as mtransforms
import matplotlib.ticker as ticker


class SquareRootScale(mscale.ScaleBase):
    """
    ScaleBase class for generating square root scale.
    """

    name = "squareroot"

    def __init__(self, axis, **kwargs):
        super().__init__(axis)

    def set_default_locators_and_formatters(self, axis):
        axis.set_major_locator(ticker.AutoLocator())
        axis.set_major_formatter(ticker.ScalarFormatter())
        axis.set_minor_locator(ticker.NullLocator())
        axis.set_minor_formatter(ticker.NullFormatter())

    def limit_range_for_scale(self, vmin, vmax, minpos):
        return max(0.0, vmin), vmax

    class SquareRootTransform(mtransforms.Transform):
        input_dims = 1
        output_dims = 1
        is_separable = True

        def transform_non_affine(self, a):
            return numpy.asarray(a) ** 0.5

        def inverted(self):
            return SquareRootScale.InvertedSquareRootTransform()

    class InvertedSquareRootTransform(mtransforms.Transform):
        input_dims = 1
        output_dims = 1
        is_separable = True

        def transform(self, a):
            return numpy.asarray(a) ** 2

        def inverted(self):
            return SquareRootScale.SquareRootTransform()

    def get_transform(self):
        return self.SquareRootTransform()


def init_matplotlib():
    # For the diagnostics widgets in headless execution
    matplotlib.use("agg")

    # Register scales
    mscale.register_scale(SquareRootScale)



from ewoksxrpd.tasks.diagnostics import DiagnoseCalibrateMultiResults
from ewoksxrpd.gui.image_viewer import ImageViewer
from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.gui.forms import input_parameters_diagnose_multicalib


__all__ = ["OWDiagnoseCalibrateMultiResults"]


class OWDiagnoseCalibrateMultiResults(
    OWTriggerWidget, ewokstaskclass=DiagnoseCalibrateMultiResults
):
    name = "DiagnoseCalibrateMultiResults"
    description = "Diagnose multi-distance calibration"
    icon = "icons/widget.png"
    want_main_area = True

    def _init_forms(self):
        parameter_info = input_parameters_diagnose_multicalib(
            self.get_default_input_values()
        )
        self._create_input_form(parameter_info)

    def _init_main_area(self):
        layout = self._get_main_layout()
        self._image_viewer = ImageViewer()
        layout.addWidget(self._image_viewer)
        super()._init_main_area()

    def _refresh_non_form_output_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_output_widgets()
            self._update_output_file()

    def _update_output_file(self):
        filename = self.get_task_input_value("filename")
        self._image_viewer.load_image(filename)



from __future__ import annotations
import h5py
import numpy


class ImageSum:
    def __init__(self, shape) -> None:
        self.shape = shape
        self.summed_image = numpy.zeros(shape)
        self.summed_monitor = None
        self.nb_images = 0

    def add_to_sum(self, image: numpy.ndarray, monitor: int | None):
        self.summed_image += image
        if monitor is not None:
            if self.summed_monitor is None:
                self.summed_monitor = 0
            self.summed_monitor += monitor
        self.nb_images += 1

    def reset(self):
        self.summed_image = numpy.zeros(self.shape)
        self.summed_monitor = None
        self.nb_images = 0


def generate_range(start: int, end_arg: int | None, nitems: int) -> range:
    end = nitems if end_arg is None else end_arg + 1

    if (end - start) > nitems:
        raise ValueError(
            f"Asked range ({start},{end}) is bigger than number of items ({nitems})"
        )

    return range(start, end)


def save_sum(
    nxdata: h5py.Group, name: str, image_sum: ImageSum
) -> tuple[h5py.Dataset, h5py.Dataset | None]:
    nxdata.create_dataset(f"{name}_nb_images", data=image_sum.nb_images)

    image_dset = nxdata.create_dataset(name, data=image_sum.summed_image)
    if "signal" not in nxdata:
        nxdata.attrs["signal"] = name

    summed_monitor = image_sum.summed_monitor
    if summed_monitor is None:
        return image_dset, None
    monitor_dset = nxdata.create_dataset(f"{name}_monitor", data=summed_monitor)
    return image_dset, monitor_dset



from contextlib import contextmanager
import logging
from typing import Dict, Mapping, Optional, Tuple
from AnyQt import QtWidgets
from ewoksorange.bindings import OWEwoksWidgetOneThread
from ewoksorange.bindings import ow_build_opts
from ewoksorange.gui.parameterform import ParameterForm


logger = logging.getLogger(__name__)


class OWTriggerWidget(OWEwoksWidgetOneThread, **ow_build_opts):
    def __init__(self, *args, **kwargs) -> None:
        self._input_form: Optional[ParameterForm] = None
        self._output_form: Optional[ParameterForm] = None
        super().__init__(*args, **kwargs)
        self._init_ui()

    def _init_ui(self):
        """Create widgets for input and output."""
        self._init_forms()
        self._init_control_area()
        self._add_input_form_widget()
        self._init_main_area()
        self._add_output_form_widget()
        self._refresh_non_form_input_widgets()
        self._refresh_non_form_output_widgets()

    def _init_forms(self) -> None:
        pass

    def _init_control_area(self) -> None:
        """Buttons to trigger execution and refresh."""
        super()._init_control_area()
        layout = self._get_control_layout()
        refresh = QtWidgets.QPushButton("Refresh")
        layout.addWidget(refresh)
        refresh.released.connect(self._refresh_widgets)

    def _add_input_form_widget(self) -> None:
        if self._input_form is not None:
            layout = self._get_control_layout()
            layout.addWidget(self._input_form)

    def _add_output_form_widget(self) -> None:
        if self._output_form is not None:
            layout = self._get_main_layout()
            layout.addWidget(self._output_form)

    def task_output_changed(self) -> None:
        self._refresh_output_widgets()
        super().task_output_changed()

    def handleNewSignals(self) -> None:
        self._refresh_input_widgets()
        super().handleNewSignals()

    def _create_input_form(self, parameter_info: dict) -> None:
        assert self._input_form is None
        form = ParameterForm(self.controlArea)
        for name, info in parameter_info.items():
            form.addParameter(
                name, **info, value_change_callback=self._input_form_edited
            )
        self._input_form = form
        self._refresh_input_form()

    def _create_output_form(self, parameter_info: dict) -> None:
        assert self._output_form is None
        form = ParameterForm(self.mainArea)
        for name, info in parameter_info.items():
            form.addParameter(name, **info)
        self._output_form = form
        self._refresh_output_form()

    def _refresh_widgets(self) -> None:
        self._refresh_input_widgets()
        self._refresh_output_widgets()

    def _refresh_input_widgets(self) -> None:
        self._refresh_input_form()
        self._refresh_non_form_input_widgets()

    def _refresh_output_widgets(self) -> None:
        self._refresh_output_form()
        self._refresh_non_form_output_widgets()

    def _refresh_non_form_input_widgets(self) -> None:
        pass

    def _refresh_non_form_output_widgets(self) -> None:
        pass

    def _refresh_input_form(self) -> None:
        """Set form values and disable rows with values from previous tasks"""
        if self._input_form is None:
            return

        # Set form value to default or dynamic inputs
        values, checked = self._values_to_form(self.get_task_input_values())
        self._input_form.set_parameter_values(values)
        self._input_form.set_parameters_checked(checked)

        # Disable form parameters with dynamic inputs
        disabled_names = self.get_dynamic_input_names()
        enabled = self._enabled_to_form(
            {name: name not in disabled_names for name in self.get_input_names()}
        )
        self._input_form.set_parameters_enabled(enabled)

    def _refresh_output_form(self) -> None:
        """Set form values"""
        if self._output_form is None:
            return
        # Set form value to task outputs
        values, checked = self._values_to_form(self.get_task_output_values())
        self._output_form.set_parameter_values(values)
        self._output_form.set_parameters_checked(checked)

    def _input_form_edited(self) -> None:
        """Store enabled form values as default inputs"""
        values = self._input_form.get_parameter_values()
        enabled = self._input_form.get_parameters_enabled()
        checked = self._input_form.get_parameters_checked()
        values = {k: v for k, v in values.items() if enabled[k]}
        parameters = self._values_from_form(values, checked)
        self.update_default_inputs(**parameters)

    def _values_from_form(
        self, values: Mapping, checked: Dict[str, bool], output: bool = False
    ) -> Mapping:
        return values

    def _values_to_form(
        self, values: Mapping, output: bool = False
    ) -> Tuple[Mapping, Dict[str, bool]]:
        return values, dict()

    def _enabled_to_form(
        self, enabled: Dict[str, bool], output: bool = False
    ) -> Dict[str, bool]:
        return enabled

    @contextmanager
    def _capture_errors(self, msg="widget update failed"):
        try:
            yield
        except Exception:
            logger.exception(msg)



import numpy
from ewoksorange.tests.utils import execute_task

from .xrpd_theory import Measurement
from orangecontrib.ewoksxrpd.background import OWSubtractBackground


def test_background_subtraction_task(
    imageSetup1SampleA: Measurement,
    image1Setup1SampleB: Measurement,
    image2Setup1SampleB: Measurement,
):
    assert_background_subtraction(
        imageSetup1SampleA, image1Setup1SampleB, image2Setup1SampleB, None
    )


def test_background_subtraction_widget(
    imageSetup1SampleA: Measurement,
    image1Setup1SampleB: Measurement,
    image2Setup1SampleB: Measurement,
    qtapp,
):
    assert_background_subtraction(
        imageSetup1SampleA, image1Setup1SampleB, image2Setup1SampleB, qtapp
    )


def assert_background_subtraction(
    imageSetup1SampleA: Measurement,
    image1Setup1SampleB: Measurement,
    image2Setup1SampleB: Measurement,
    qtapp,
):
    image = imageSetup1SampleA.image + image1Setup1SampleB.image
    inputs = {
        "image": image,
        "monitor": imageSetup1SampleA.monitor,
        "background": image2Setup1SampleB.image,
        "background_monitor": image2Setup1SampleB.monitor,
    }
    results = execute_task(
        OWSubtractBackground.ewokstaskclass if qtapp is None else OWSubtractBackground,
        inputs=inputs,
    )
    numpy.testing.assert_allclose(
        imageSetup1SampleA.image, results["image"], atol=1e-10
    )



import json
from numbers import Integral
from typing import Union
import numpy


def json_dumps(value) -> str:
    return json.dumps(value)


def json_loads(value: str):
    return json.loads(value)


def float_serialize(value: float) -> str:
    return str(value)


def float_deserialize(value: str) -> float:
    return float(value)


def strfloat_serialize(value: Union[float, str]) -> str:
    return str(value)


def strfloat_deserialize(value: str) -> float:
    return float(value)


def cm_serialize(value: float) -> str:
    return str(value * 1e2)


def cm_deserialize(value: str) -> float:
    return float(value) * 1e-2


def degrees_serialize(value: float) -> str:
    return str(numpy.degrees(value))


def degrees_deserialize(value: str) -> float:
    return numpy.radians(float(value))


def posint_serialize(value: Integral) -> Integral:
    return max(value, 0)


def posint_deserialize(value: Integral) -> Integral:
    return max(value, 0)



import os
from numbers import Integral
import pytest

from ewoksorange.tests.utils import execute_task
from ewoksxrpd.tasks import sum


@pytest.mark.parametrize("monitor_name", ("mon", None))
@pytest.mark.parametrize("background_step", [False, True])
def test_sum_bliss_scan(tmpdir, bliss_perkinelmer_scan, monitor_name, background_step):
    inputs = {
        "filename": str(bliss_perkinelmer_scan),
        "detector_name": "perkinelmer",
        "monitor_name": monitor_name,
        "output_filename": str(tmpdir / "output.h5"),
        "scan": 2,
        "background_step": background_step,
    }

    outputs = execute_task(sum.SumBlissScanImages, inputs=inputs)

    assert os.path.isfile(str(tmpdir / "output.h5"))
    assert isinstance(outputs["output_uri"], str)
    if monitor_name is None:
        assert outputs["monitor"] is None
    else:
        assert isinstance(outputs["monitor"], Integral)


@pytest.mark.parametrize("monitor_name", ("mon", None))
def test_sum(tmpdir, bliss_perkinelmer_scan, monitor_name):
    inputs = {
        "filename": str(bliss_perkinelmer_scan),
        "detector_name": "perkinelmer",
        "monitor_name": monitor_name,
        "output_filename": str(tmpdir / "output.h5"),
        "start_scan": 2,
        "end_image": 10,
        "end_scan": 2,
    }

    outputs = execute_task(sum.SumImages, inputs=inputs)

    assert os.path.isfile(str(tmpdir / "output.h5"))
    assert len(outputs["output_uris"]) == 1
    if monitor_name is None:
        assert outputs["monitor_uris"] is None
    else:
        assert len(outputs["monitor_uris"]) == 1


def test_sum_type_both(tmpdir, bliss_perkinelmer_scan):
    inputs = {
        "filename": str(bliss_perkinelmer_scan),
        "detector_name": "perkinelmer",
        "output_filename": str(tmpdir / "output.h5"),
        "start_scan": 2,
        "end_scan": 2,
        "sum_type": "both",
        "monitor_name": "mon",
    }

    outputs = execute_task(
        sum.SumImages,
        inputs=inputs,
    )

    assert os.path.isfile(str(tmpdir / "output.h5"))
    # Scan sum and full sum
    assert len(outputs["output_uris"]) == 2
    assert len(outputs["monitor_uris"]) == 2


@pytest.mark.parametrize("background_step", range(-1, 4))
def test_tscan_background_step(background_step):
    for npoints in range(1, 21):
        total_points, background_step, include_indices, skip_indices = tscan(
            npoints, background_step=background_step
        )
        assert npoints == len(include_indices)
        assert total_points == len(include_indices) + len(skip_indices)

        counter_iterator = range(total_points)
        include_indices2 = [
            i for i, _ in sum.iterate_scan_with_skip(counter_iterator, background_step)
        ]

        assert include_indices == include_indices2


def tscan(npoints: int, background_step: int):
    """
    :param npoints: number of lima images
    :param background_step: -1 means no background, 0 means only one dark, N>0 means a dark every N image
    """

    # Keep this logic as much as possible, even when we can make it cleaner:
    # https://gitlab.esrf.fr/bcu-vercors/id22/id22/-/blob/master/id22/scripts/tscan.py

    last_image = -1
    if background_step > npoints:
        background_step = npoints
    if background_step >= 0:
        if background_step == 0:
            bkg_range = [0]
        else:
            bkg_range = range(0, npoints, background_step)

        # now calculate the real range for bkgd image in total scan
        nb_background = len(bkg_range)
        total_points = npoints + nb_background
        last_image = total_points - 1

        if background_step > 0:
            bkg_range = range(0, total_points, background_step + 1)
    else:
        total_points = npoints
        bkg_range = None

    save_manual = bkg_range is not None  # this is missing in the original code!

    shutter = "OPEN"
    open_indices = list()
    closed_indices = list()
    for iteration in range(total_points):
        # TscanChainPreset.Iterator.start
        if bkg_range and iteration in bkg_range:
            shutter = "CLOSED"

        # Measure scan point `iteration`
        if shutter == "CLOSED":
            closed_indices.append(iteration)
        else:
            assert shutter == "OPEN"
            if not save_manual:
                open_indices.append(iteration)

        # TscanChainPreset.Iterator.stop
        if save_manual and bkg_range is not None:
            if (
                bkg_range is not None
                and iteration not in bkg_range
                and iteration != last_image
            ):
                assert shutter == "OPEN"
                open_indices.append(iteration)
        if bkg_range and iteration in bkg_range:
            shutter = "OPEN"

    # TscanChainPreset.before_stop
    if last_image >= 0:
        open_indices.append(iteration)

    # TscanChainPreset.stop
    shutter = "CLOSED"

    return total_points, background_step, open_indices, closed_indices



from typing import Dict, Mapping, Optional, Sequence, Tuple
from . import serialize

from pyFAI.detectors import ALL_DETECTORS
from pyFAI.calibrant import ALL_CALIBRANTS
from ewokscore import missing_data


_GEOMETRY_KEYS = "dist", "poni1", "poni2", "rot1", "rot2", "rot3"
_ENERGY_GEOMETRY_KEYS = ("energy",) + _GEOMETRY_KEYS
_PARAMETRIZATION_KEYS = (
    "dist_expr",
    "param_names",
    "poni1_expr",
    "poni2_expr",
    "pos_names",
    "rot1_expr",
    "rot2_expr",
    "rot3_expr",
    "wavelength_expr",
)
_PARAMETERS_KEYS = (
    "arot1",
    "arot2",
    "arot3",
    "dist_offset",
    "dist_scale",
    "energy",
    "poni1_offset",
    "poni1_scale",
    "poni2_offset",
    "poni2_scale",
)


def input_parameters_calibratesingle(values: Mapping) -> dict:
    parameters = dict()
    _add_data_singlecalib(parameters)
    _add_energy_geometry(parameters, values, calib=True)
    _add_detector_input(parameters)
    parameters["ring_detector"] = {
        "label": "Detector for the ring extraction. Leave this empty to use the same detector.",
        "value_for_type": list(ALL_DETECTORS),
        "serialize": str.lower,
    }
    _add_calibrant_input(parameters)
    _apply_values(parameters, values)
    return parameters


def output_parameters_calibratesingle() -> dict:
    parameters = dict()
    _add_energy_geometry(parameters, dict())
    return parameters


def input_parameters_calibratemulti(values: Mapping) -> dict:
    parameters = dict()
    _add_data_multicalib(parameters)
    _add_energy_geometry(parameters, values, calib=True)
    _add_detector_input(parameters)
    _add_calibrant_input(parameters)
    _apply_values(parameters, values)
    return parameters


def output_parameters_calibratemulti() -> dict:
    parameters = dict()
    _add_parametrization(parameters, dict())
    return parameters


def input_parameters_integrate1d(values: Mapping) -> dict:
    parameters = dict()
    _add_data_integrate1d(parameters)
    _add_energy_geometry(parameters, values)
    _add_detector_input(parameters)
    parameters["integration_options"] = {
        "label": "Integrate Options",
        "value_for_type": "",
        "serialize": serialize.json_dumps,
        "deserialize": serialize.json_loads,
    }
    parameters["fixed_integration_options"] = {
        "label": "Fixed Integrate Options",
        "value_for_type": "",
        "serialize": serialize.json_dumps,
        "deserialize": serialize.json_loads,
    }
    _apply_values(parameters, values)
    return parameters


def input_parameters_integrateblissscan(values: Mapping, saving: bool = False) -> dict:
    parameters = dict()
    _add_data_integrateblissscan(parameters)
    _add_energy_geometry(parameters, values)
    _add_detector_input(parameters)
    parameters["integration_options"] = {
        "label": "Integrate Options",
        "value_for_type": "",
        "serialize": serialize.json_dumps,
        "deserialize": serialize.json_loads,
    }
    parameters["fixed_integration_options"] = {
        "label": "Fixed Integrate Options",
        "value_for_type": "",
        "serialize": serialize.json_dumps,
        "deserialize": serialize.json_loads,
    }
    if saving:
        parameters["output_filename"] = {
            "label": "Output filename",
            "value_for_type": "",
            "select": "file",
        }
        parameters["nxprocess_as_default"] = {
            "label": "Mark Result as Default",
            "value_for_type": False,
        }
    _apply_values(parameters, values)
    return parameters


def input_parameters_ascii(values: Mapping) -> dict:
    parameters = {
        "filename": {"label": "Output file", "value_for_type": "", "select": "newfile"}
    }
    _apply_values(parameters, values)
    return parameters


def input_parameters_nexus(values: Mapping) -> dict:
    parameters = {
        "url": {"label": "Output file", "value_for_type": "", "select": "h5group"},
        "nxprocess_as_default": {
            "label": "Mark Result as Default",
            "value_for_type": False,
        },
    }
    _apply_values(parameters, values)
    return parameters


def input_parameters_background(values: Mapping) -> dict:
    parameters = {
        "image": {"label": "Image", "value_for_type": "", "select": "h5dataset"},
        "monitor": {"label": "Monitor", "value_for_type": "", "select": "h5dataset"},
        "background": {
            "label": "Background",
            "value_for_type": "",
            "select": "h5dataset",
        },
        "background_monitor": {
            "label": "Background Monitor",
            "value_for_type": "",
            "select": "h5dataset",
        },
    }
    _apply_values(parameters, values)
    return parameters


def input_parameters_mask(values: Mapping) -> dict:
    parameters = {
        "image1": {"label": "Image 1", "value_for_type": "", "select": "h5dataset"},
        "monitor1": {"label": "Monitor 1", "value_for_type": "", "select": "h5dataset"},
        "image2": {"label": "Image 2", "value_for_type": "", "select": "h5dataset"},
        "monitor2": {"label": "Monitor 2", "value_for_type": "", "select": "h5dataset"},
        "smooth": {
            "label": "Smooth width",
            "value_for_type": 0,
        },
        "monitor_ratio_margin": {
            "label": "Ratio margin",
            "value_for_type": "",
            "serialize": serialize.float_serialize,
            "deserialize": serialize.float_deserialize,
        },
    }
    _apply_values(parameters, values)
    return parameters


def input_parameters_calculategeometry(values: Mapping) -> dict:
    parameters = {
        "position": {"label": "Position", "value_for_type": "", "select": "h5dataset"},
    }
    _apply_values(parameters, values)
    return parameters


def output_parameters_calculategeometry() -> dict:
    parameters = dict()
    _add_energy_geometry(parameters, dict())
    return parameters


def input_parameters_diagnose_integrate1d(values: Mapping) -> dict:
    parameters = dict()
    _add_calibrant_input(parameters)
    _add_diagnostics_output(parameters)
    _apply_values(parameters, values)
    return parameters


def input_parameters_diagnose_singlecalib(values: Mapping) -> dict:
    parameters = dict()
    _add_data_singlecalib(parameters)
    _add_calibrant_input(parameters)
    _add_diagnostics_output(parameters)
    _apply_values(parameters, values)
    return parameters


def input_parameters_diagnose_multicalib(values: Mapping) -> dict:
    parameters = dict()
    _add_diagnostics_output(parameters)
    parameters["positions"] = {
        "label": "Positions",
        "select": "h5datasets",
        "value_for_type": "",
        "serialize": serialize.json_dumps,
        "deserialize": serialize.json_loads,
    }
    parameters["positionunits_in_meter"] = {
        "label": "Position Units (meter)",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    _apply_values(parameters, values)
    return parameters


def input_parameters_pyfai_config(values: Mapping) -> dict:
    parameters = dict()
    parameters["filename"] = {
        "label": "Poni file",
        "value_for_type": "",
        "select": "file",
    }
    _add_energy_geometry(parameters, dict())
    _add_calibrant_input(parameters)
    _add_detector_input(parameters)
    parameters["mask"] = {
        "label": "Mask file",
        "value_for_type": "",
        "select": "file",
    }
    parameters["darkflatmethod"] = {
        "label": "Dark-Flat Method",
        "value_for_type": ["counts"],
        "select": "choices",
    }
    parameters["integration_options"] = {
        "label": "Integrate Options",
        "value_for_type": "",
        "serialize": serialize.json_dumps,
        "deserialize": serialize.json_loads,
    }
    _apply_values(parameters, values)
    return parameters


def output_parameters_pyfai_config() -> dict:
    parameters = dict()
    _add_energy_geometry(parameters, dict())
    _add_detector_output(parameters)
    _add_calibrant_output(parameters)
    return parameters


def _add_diagnostics_output(parameters: dict) -> None:
    parameters["filename"] = {
        "label": "Output file",
        "value_for_type": "",
        "select": "newfile",
    }


def _add_energy_geometry(
    parameters: dict, values: Mapping, calib: Optional[bool] = None
) -> None:
    values = _unpack(values, _GEOMETRY_KEYS, "geometry")
    fixed = values.pop("fixed", None)

    parameters["energy"] = {
        "label": "Energy (keV)",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["dist"] = {
        "label": "Distance (cm)",
        "value_for_type": "",
        "serialize": serialize.cm_serialize,
        "deserialize": serialize.cm_deserialize,
    }
    parameters["poni1"] = {
        "label": "Poni1 (cm)",
        "value_for_type": "",
        "serialize": serialize.cm_serialize,
        "deserialize": serialize.cm_deserialize,
    }
    parameters["poni2"] = {
        "label": "Poni2 (cm)",
        "value_for_type": "",
        "serialize": serialize.cm_serialize,
        "deserialize": serialize.cm_deserialize,
    }
    parameters["rot1"] = {
        "label": "Rot1 (deg)",
        "value_for_type": "",
        "serialize": serialize.degrees_serialize,
        "deserialize": serialize.degrees_deserialize,
    }
    parameters["rot2"] = {
        "label": "Rot2 (deg)",
        "value_for_type": "",
        "serialize": serialize.degrees_serialize,
        "deserialize": serialize.degrees_deserialize,
    }
    parameters["rot3"] = {
        "label": "Rot3 (deg)",
        "value_for_type": "",
        "serialize": serialize.degrees_serialize,
        "deserialize": serialize.degrees_deserialize,
    }

    if calib:
        if not fixed:
            fixed = list()
        for name in _ENERGY_GEOMETRY_KEYS:
            parameters[name].update(
                {
                    "checked": name not in fixed,
                    "checkbox_label": "refine",
                }
            )
        parameters["max_rings"] = {
            "label": "# Rings",
            "value_for_type": 0,
            "serialize": serialize.posint_serialize,
            "deserialize": serialize.posint_deserialize,
        }
        parameters["robust"] = {
            "label": "Robust",
            "value_for_type": True,
        }


def _add_parametrization(parameters: dict, values: Mapping) -> None:
    values = _unpack(values, _PARAMETRIZATION_KEYS, "parametrization")
    values = _unpack(values, _PARAMETERS_KEYS, "parameters")

    parameters["energy"] = {
        "label": "energy (keV)",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["wavelength_expr"] = {
        "label": "wavelength (m) =",
        "value_for_type": "",
    }

    parameters["dist_offset"] = {
        "label": "dist_offset (cm)",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["dist_scale"] = {
        "label": "dist_scal",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["dist_expr"] = {
        "label": "dist (m) =",
        "value_for_type": "",
    }

    parameters["poni1_offset"] = {
        "label": "poni1_offset (cm)",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["poni1_scale"] = {
        "label": "poni1_scale",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["poni1_expr"] = {
        "label": "poni1 (m) =",
        "value_for_type": "",
    }

    parameters["poni2_offset"] = {
        "label": "poni2_offset (cm)",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["poni2_scale"] = {
        "label": "poni2_scale",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["poni2_expr"] = {
        "label": "poni2 (m) =",
        "value_for_type": "",
    }

    parameters["arot1"] = {
        "label": "arot1 (rad)",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["rot1_expr"] = {
        "label": "rot1 (rad) =",
        "value_for_type": "",
    }

    parameters["arot2"] = {
        "label": "arot2 (rad)",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["rot2_expr"] = {
        "label": "rot2 (rad) =",
        "value_for_type": "",
    }

    parameters["arot3"] = {
        "label": "arot3 (drad)",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["rot3_expr"] = {
        "label": "rot3 (rad) =",
        "value_for_type": "",
    }


def _add_detector_input(parameters: dict) -> None:
    parameters["detector"] = {
        "label": "Detector",
        "value_for_type": list(ALL_DETECTORS),
        "serialize": str.lower,
    }
    parameters["detector_config"] = {
        "label": "Detector parameters",
        "value_for_type": "",
        "serialize": serialize.json_dumps,
        "deserialize": serialize.json_loads,
    }


def _add_detector_output(parameters: dict) -> None:
    parameters["detector"] = {
        "label": "Detector",
        "value_for_type": "",
        "serialize": str.lower,
    }
    parameters["detector_config"] = {
        "label": "Detector parameters",
        "value_for_type": "",
        "serialize": serialize.json_dumps,
        "deserialize": serialize.json_loads,
    }


def _add_calibrant_input(parameters: dict) -> None:
    parameters["calibrant"] = {
        "label": "Calibrant",
        "value_for_type": list(ALL_CALIBRANTS.all),
    }


def _add_calibrant_output(parameters: dict) -> None:
    parameters["calibrant"] = {
        "label": "Calibrant",
        "value_for_type": "",
    }


def output_parameters_calib(parameters: dict):
    _add_energy_geometry(parameters, dict(), False)


def _add_data_singlecalib(parameters: dict) -> None:
    parameters["image"] = {
        "label": "Pattern",
        "value_for_type": "",
        "select": "h5dataset",
    }


def _add_data_multicalib(parameters: dict) -> None:
    parameters["images"] = {
        "label": "Patterns",
        "value_for_type": "",
        "select": "h5datasets",
        "serialize": serialize.json_dumps,
        "deserialize": serialize.json_loads,
    }
    parameters["positions"] = {
        "label": "Positions",
        "value_for_type": "",
        "select": "h5datasets",
        "serialize": serialize.json_dumps,
        "deserialize": serialize.json_loads,
    }
    parameters["reference_position"] = {
        "label": "Reference Position",
        "value_for_type": "",
        "select": "h5dataset",
    }
    parameters["sample_position"] = {
        "label": "Sample position",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["positionunits_in_meter"] = {
        "label": "Position Units (meter)",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }


def _add_data_integrate1d(parameters: dict) -> None:
    parameters["image"] = {
        "label": "Pattern",
        "value_for_type": "",
        "select": "h5dataset",
    }
    parameters["monitor"] = {
        "label": "Monitor",
        "value_for_type": "",
        "select": "h5dataset",
    }
    parameters["reference"] = {
        "label": "Reference",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["mask"] = {"label": "Mask", "value_for_type": "", "select": "h5dataset"}
    _add_detector_input(parameters)


def _add_data_integrateblissscan(parameters: dict) -> None:
    parameters["filename"] = {
        "label": "Bliss filename",
        "value_for_type": "",
        "select": "file",
    }
    parameters["scan"] = {"label": "Scan number", "value_for_type": 0}
    parameters["subscan"] = {"label": "Sub scan number", "value_for_type": 0}

    parameters["detector_name"] = {"label": "Detector name", "value_for_type": ""}
    parameters["counter_names"] = {
        "label": "Counter names",
        "value_for_type": "",
        "serialize": serialize.json_dumps,
        "deserialize": serialize.json_loads,
    }
    parameters["monitor"] = {
        "label": "Monitor",
        "value_for_type": "",
        "select": "h5dataset",
    }
    parameters["reference"] = {
        "label": "Reference",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["mask"] = {"label": "Mask", "value_for_type": "", "select": "h5dataset"}
    _add_detector_input(parameters)

    parameters["retry_timeout"] = {
        "label": "retry_timeout",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }

    parameters["retry_period"] = {
        "label": "retry_period",
        "value_for_type": "",
        "serialize": serialize.float_serialize,
        "deserialize": serialize.float_deserialize,
    }
    parameters["demo"] = {"label": "demo", "value_for_type": False}
    return parameters


def _apply_values(parameters: dict, values: Mapping) -> None:
    for k, v in values.items():
        if k in parameters:
            parameters[k]["value"] = v


def unpack_geometry(values: Mapping) -> Tuple[Mapping, Dict[str, bool]]:
    values = _unpack(values, _GEOMETRY_KEYS, "geometry")
    fixed = values.pop("fixed", None)
    if fixed:
        checked = {name: name not in fixed for name in _ENERGY_GEOMETRY_KEYS}
    else:
        checked = dict()
    return values, checked


def unpack_enabled_geometry(enabled: Dict[str, bool]) -> Dict[str, bool]:
    values = _unpack_enabled(enabled, _GEOMETRY_KEYS, "geometry")
    return values


def pack_geometry(values: Mapping, checked: Dict[str, bool]) -> dict:
    values = _pack(values, _GEOMETRY_KEYS, "geometry")
    if checked:
        values["fixed"] = [k for k, v in checked.items() if not v]
    else:
        values["fixed"] = missing_data.MISSING_DATA
    return values


def unpack_parametrization(values: Mapping) -> dict:
    values = _unpack(values, _PARAMETRIZATION_KEYS, "parametrization")
    values = _unpack(values, _PARAMETERS_KEYS, "parameters")
    return values


def unpack_enabled_parametrization(enabled: Dict[str, bool]) -> Dict[str, bool]:
    values = _unpack_enabled(enabled, _PARAMETRIZATION_KEYS, "parametrization")
    values = _unpack_enabled(enabled, _PARAMETERS_KEYS, "parameters")
    return values


def pack_parametrization(values: Mapping) -> dict:
    values = _pack(values, _PARAMETRIZATION_KEYS, "parametrization")
    values = _pack(values, _PARAMETERS_KEYS, "parameters")
    return values


def _pack(values: Mapping, keys: Sequence[str], pack_key: str) -> dict:
    result = {k: v for k, v in values.items() if k not in keys}
    packed = {
        k: v
        for k, v in values.items()
        if k in keys and not missing_data.is_missing_data(v)
    }
    if len(packed) == len(keys):
        result[pack_key] = packed
    else:
        result[pack_key] = missing_data.MISSING_DATA
    return result


def _unpack(values: Mapping, keys: Sequence[str], pack_key: str) -> dict:
    result = dict(values)
    packed = result.pop(pack_key, None)
    if packed:
        result.update(packed)
    else:
        result.update({k: missing_data.MISSING_DATA for k in keys})
    return result


def _unpack_enabled(
    enabled: Dict[str, bool], keys: Sequence[str], pack_key: str
) -> Dict[str, bool]:
    result = dict(enabled)
    packed = result.pop(pack_key, None)
    if packed is not None:
        result.update({k: packed for k in keys})
    return result



from typing import Dict, Iterable, List, Mapping, Tuple
from AnyQt import QtWidgets
from silx.gui.plot import ScatterView, PlotWidget
from ewokscore.missing_data import is_missing_data
from ewoksdata.data import bliss

from ewoksxrpd.tasks.calibrate import CalibrateMulti
from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.gui import plots
from ewoksxrpd.tasks.utils import data_utils
from ewoksxrpd.tasks.calibrate import calculate_geometry
from ewoksxrpd.gui.forms import input_parameters_calibratemulti
from ewoksxrpd.gui.forms import output_parameters_calibratemulti
from ewoksxrpd.gui.forms import pack_geometry
from ewoksxrpd.gui.forms import unpack_geometry
from ewoksxrpd.gui.forms import unpack_enabled_geometry
from ewoksxrpd.gui.forms import pack_parametrization
from ewoksxrpd.gui.forms import unpack_parametrization
from ewoksxrpd.gui.forms import unpack_enabled_parametrization

__all__ = ["OWCalibrateMulti"]


class OWCalibrateMulti(OWTriggerWidget, ewokstaskclass=CalibrateMulti):
    name = "CalibrateMulti"
    description = "Multi-distance calibration"
    icon = "icons/widget.png"
    want_main_area = True

    def __init__(self, *args, **kwargs) -> None:
        self._tabs = QtWidgets.QTabWidget()
        self._show = {
            "output_rings": True,
            "detected_rings": False,
        }
        self._legends = list()
        self._cache = list()
        self._fixed_tabs = 0
        super().__init__(*args, **kwargs)

    def _init_forms(self):
        parameter_info = input_parameters_calibratemulti(
            self.get_default_input_values()
        )
        self._create_input_form(parameter_info)
        parameter_info = output_parameters_calibratemulti()
        self._create_output_form(parameter_info)

    def _init_control_area(self):
        super()._init_control_area()
        layout = self._get_control_layout()

        w = QtWidgets.QPushButton("Accept refined")
        layout.addWidget(w)
        w.released.connect(self._accept_refined_parameters)

        self._show_output_rings_widget = w = QtWidgets.QCheckBox("Refined rings")
        w.setChecked(self._show["output_rings"])
        layout.addWidget(w)
        w.released.connect(self._set_show_output_rings)

        self._show_detected_rings_widget = w = QtWidgets.QCheckBox("Detected rings")
        w.setChecked(self._show["detected_rings"])
        layout.addWidget(w)
        w.released.connect(self._set_show_detected_rings)

    def _set_show_output_rings(self):
        self._show["output_rings"] = self._show_output_rings_widget.isChecked()
        self._refresh_non_form_output_widgets()

    def _set_show_detected_rings(self):
        self._show["detected_rings"] = self._show_detected_rings_widget.isChecked()
        self._refresh_non_form_output_widgets()

    def _init_main_area(self):
        layout = self._get_main_layout()
        layout.addWidget(self._tabs)
        self._fixed_tabs = 1
        self._tabs.addTab(self._output_form, "Parametrization")
        super()._init_main_area()
        self._refresh_non_form_input_widgets()
        self._refresh_non_form_output_widgets()

    def _add_output_form_widget(self):
        pass

    def _refresh_non_form_input_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_input_widgets()
            self._refresh_input_plots()

    def _refresh_non_form_output_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_output_widgets()
            self._refresh_output_plots()

    def _accept_refined_parameters(self):
        parametrization = self.get_task_output_value("parametrization")
        parameters = self.get_task_output_value("parameters")
        position = self.get_task_input_value("reference_position")
        if not parametrization or not parameters:
            return
        if is_missing_data(position):
            position = 0
        else:
            position = bliss.get_data(position, retry_timeout=0)
        parametrization = data_utils.data_from_storage(parametrization)
        parameters = data_utils.data_from_storage(parameters)
        geometry, energy = calculate_geometry(parametrization, parameters, position)
        self.update_default_inputs(energy=energy, geometry=geometry)
        self._refresh_input_widgets()

    def _input_form_edited(self):
        super()._input_form_edited()
        fixed = [
            k for k, v in self._input_form.get_parameters_checked().items() if not v
        ]
        self.update_default_inputs(fixed=fixed)
        self._refresh_non_form_input_widgets()

    def _refresh_input_plots(self):
        self._update_tabs()
        nplots = self._get_nplots()
        if not nplots:
            return None
        inputs = self.get_task_input_values()
        for plot_index in range(nplots):
            self._update_image(inputs, plot_index)

    def _refresh_output_plots(self):
        nplots = self._get_nplots()
        if not nplots:
            return None
        outputs = self.get_task_output_values()
        inoutputs = {**self.get_task_input_values(), **outputs}
        for plot_index in range(nplots):
            self._update_detected_rings(outputs, plot_index)
            self._update_output_rings(inoutputs, plot_index)

    def _update_image(self, inputs, plot_index):
        image = inputs["images"][plot_index]
        if isinstance(image, str):
            cache = self._cache[plot_index]
            previous_image_url = cache.get("image")
            if previous_image_url == image:
                return
            cache["image"] = image
        self._remove_from_plot("image", plot_index)
        if not data_utils.is_data(image):
            return
        image = bliss.get_image(image, retry_timeout=0)
        self._legends[plot_index]["image"] = [
            plots.plot_image(self._get_plot(plot_index), image, legend="image")
        ]

    def _update_output_rings(self, values, plot_index):
        self._remove_from_plot("output_rings", plot_index)
        if not self._show["output_rings"]:
            return
        self._legends[plot_index]["output_rings"] = self._update_theoretical_rings(
            values, "output", plot_index
        )

    def _update_theoretical_rings(self, values, legend, plot_index) -> List[str]:
        energy = values.get("energy")
        detector = values.get("detector")
        detector_config = values.get("detector_config")
        calibrant = values.get("calibrant")
        parametrization = values.get("parametrization")
        parameters = values.get("parameters")
        positions = values.get("positions")
        if (
            not energy
            or not detector
            or not calibrant
            or not parametrization
            or not parameters
            or not positions
        ):
            return
        position = bliss.get_data(positions[plot_index], retry_timeout=0)
        parametrization = data_utils.data_from_storage(parametrization)
        parameters = data_utils.data_from_storage(parameters)
        geometry, energy = calculate_geometry(parametrization, parameters, position)
        plot = self._get_plot(plot_index)
        return plots.plot_theoretical_rings(
            plot,
            detector,
            calibrant,
            energy,
            geometry,
            detector_config=detector_config,
            max_rings=None,
            legend=legend,
        )

    def _update_detected_rings(self, values, plot_index):
        self._remove_from_plot("detected_rings", plot_index)
        if not self._show["detected_rings"]:
            return
        rings = values.get("rings")
        if not rings:
            return
        rings = data_utils.data_from_storage(rings)
        plot = self._get_plot(plot_index)
        self._legends[plot_index]["detected_rings"] = plots.plot_detected_rings(
            plot, rings[str(plot_index)]
        )

    def _remove_from_plot(self, name: str, plot_index: int) -> None:
        legends = self._legends[plot_index].pop(name, list())
        if legends:
            plot = self._get_plot(plot_index)
            for legend in legends:
                plot.remove(legend=legend)

    def _values_from_form(
        self, values: Mapping, checked: Dict[str, bool], output: bool = False
    ) -> Mapping:
        values = pack_geometry(values, checked)
        if output:
            values = pack_parametrization(values)
        return values

    def _values_to_form(
        self, values: Mapping, output: bool = False
    ) -> Tuple[Mapping, Dict[str, bool]]:
        values, checked = unpack_geometry(values)
        if output:
            values = unpack_parametrization(values)
        return values, checked

    def _enabled_to_form(
        self, enabled: Dict[str, bool], output: bool = False
    ) -> Dict[str, bool]:
        enabled = unpack_enabled_geometry(enabled)
        if output:
            enabled = unpack_enabled_parametrization(enabled)
        return enabled

    def _get_nplots(self) -> Iterable[PlotWidget]:
        return self._tabs.count() - self._fixed_tabs

    def _get_plot(self, plot_index) -> PlotWidget:
        return self._tabs.widget(plot_index + self._fixed_tabs).getPlotWidget()

    def _update_tabs(self):
        images = self.get_task_input_value("images")
        if images:
            images = [data_utils.get_image(image, retry_timeout=0) for image in images]
            positions = self.get_task_input_value("positions")
            if positions:
                positions = [
                    data_utils.get_data(position, retry_timeout=0)
                    for position in positions
                ]
            else:
                positions = list()
            positions += [float("nan")] * max(len(images) - len(positions), 0)
        else:
            images = list()
            positions = list()

        xunits_in_m = self.get_task_input_value("positionunits_in_meter")  # xunits/m
        if not xunits_in_m:
            xunits_in_m = 1e-3  # xunits/m
        dunits_in_m = 1e-2  # dunits/m
        positions = [
            f"Position = {position*xunits_in_m/dunits_in_m:.06f} cm"
            for position in positions
        ]

        ntabs = self._tabs.count()
        nfixed = self._fixed_tabs
        nplots = ntabs - nfixed
        nadd = len(images) - nplots

        if nadd == 0:
            pass
        elif nadd > 0:
            for tab_index in range(ntabs, ntabs + nadd):
                plot_index = tab_index - nfixed
                self._tabs.addTab(ScatterView(), positions[plot_index])
                self._legends.append(dict())
                self._cache.append(dict())
        else:
            for tab_index in range(ntabs - 1, ntabs + nadd - 1, -1):
                plot_index = tab_index - nfixed
                self._tabs.removeTab(tab_index)
                del self._legends[plot_index]
                del self._cache[plot_index]

        for plot_index, position in enumerate(positions):
            tab_index = plot_index + nfixed
            self._tabs.setTabText(tab_index, position)



import time
import logging
from typing import Iterator, Optional, Any, Tuple, Generator

_logger = logging.getLogger(__name__)


def zip_with_progress(
    *iterators: Iterator[Iterator[Any]],
    message: str = "Progress %d/%s",
    nmax: Optional[int] = None,
    period: float = 5,
    logger=None,
) -> Generator[Tuple[Any], None, None]:
    """Like python's zip but will progress logging when iterating over the result."""
    if logger is None:
        logger = _logger
    i = 0
    t0 = time.time()
    if nmax is None:
        nmax = "?"
    try:
        for tpl in zip(*iterators):
            yield tpl
            i += 1
            if (time.time() - t0) > period:
                t0 = time.time()
                _logger.info(message, i, nmax)
    finally:
        _logger.info(f"{message} (FINISHED)", i, nmax)



from ..tasks.utils import pyfai_utils


def test_normalize_parameters():
    parameters = {
        "key1": 10,
        "key2": "normal",
        "key3": [
            "/gpfs/jazzy",
            "/gpfs/jazzy/",
            "silx:///gpfs/jazzy/path/to/file",
            "/mnt/multipath-shares/path/to/file",
        ],
    }
    parameters["key4"] = {
        "key1": 10,
        "key2": "normal",
        "key3": [
            "/gpfs/jazzy",
            "/gpfs/jazzy/",
            "silx:///gpfs/jazzy/path/to/file",
            "/mnt/multipath-shares/path/to/file",
        ],
    }
    normalized = {
        "key1": 10,
        "key2": "normal",
        "key3": ["/gpfs/jazzy", "/", "silx:///path/to/file", "/path/to/file"],
    }
    normalized["key4"] = {
        "key1": 10,
        "key2": "normal",
        "key3": ["/gpfs/jazzy", "/", "silx:///path/to/file", "/path/to/file"],
    }
    assert pyfai_utils.normalize_parameters(parameters) == normalized



import logging
from pyFAI import method_registry
from ewokscore import Task

from .utils import pyfai_utils, xrpd_utils

__all__ = ["PyFaiConfig"]

logger = logging.getLogger(__name__)


class PyFaiConfig(
    Task,
    optional_input_names=[
        "filename",
        "filenames",
        "energy",
        "geometry",
        "mask",
        "flatfield",
        "darkcurrent",
        "darkflatmethod",
        "detector",
        "detector_config",
        "calibrant",
        "integration_options",
    ],
    output_names=[
        "energy",
        "geometry",
        "detector",
        "detector_config",
        "calibrant",
        "mask",
        "flatfield",
        "darkcurrent",
        "integration_options",
    ],
):
    """Parse pyFAI calibration and integration parameters"""

    def run(self):
        input_values = self.input_values
        integration_options = self.merged_integration_options()

        wavelength = integration_options.pop("wavelength", None)
        energy = input_values.get("energy", None) or integration_options.pop(
            "energy", None
        )
        if energy is None:
            if wavelength is not None:
                energy = xrpd_utils.energy_wavelength(wavelength)

        detector = integration_options.pop("detector", None)
        if not self.missing_inputs.detector:
            detector = input_values["detector"]

        detector_config = integration_options.pop("detector_config", None)
        if not self.missing_inputs.detector_config:
            detector_config = input_values["detector_config"]

        calibrant = input_values.get("calibrant", None)

        mask = input_values.get("mask", None)
        flatfield = input_values.get("flatfield", None)
        darkcurrent = input_values.get("darkcurrent", None)
        if not self.missing_inputs.darkflatmethod:
            integration_options["darkflatmethod"] = self.inputs.darkflatmethod

        geometry = {
            k: integration_options.pop(k)
            for k in ["dist", "poni1", "poni2", "rot1", "rot2", "rot3"]
            if k in integration_options
        }
        if not self.missing_inputs.geometry:
            geometry = input_values["geometry"]

        do_poisson = integration_options.pop("do_poisson", None)
        do_azimuthal_error = integration_options.pop("do_azimuthal_error", None)
        error_model = integration_options.pop("error_model", None)
        if not error_model:
            if do_poisson:
                error_model = "poisson"
            if do_azimuthal_error:
                error_model = "azimuthal"
        if error_model:
            integration_options["error_model"] = error_model

        # Check method and integrator function
        method = integration_options.get("method", "")
        if not isinstance(method, str):
            method = "_".join(method)
        pmethod = method_registry.Method.parsed(method)

        integrator_name = integration_options.get("integrator_name", "")
        if integrator_name in ("sigma_clip", "_sigma_clip_legacy"):
            logger.warning(
                "'%s' is not compatible with the pyfai worker: use 'sigma_clip_ng'",
                integrator_name,
            )
            integration_options["integrator_name"] = "sigma_clip_ng"
        if "sigma_clip_ng" == integrator_name and pmethod.split != "no":
            raise ValueError(
                "to combine sigma clipping with pixel splitting, use 'sigma_clip_legacy'"
            )

        # Split integration and worker options
        self.outputs.energy = energy
        self.outputs.geometry = geometry
        self.outputs.detector = detector
        self.outputs.detector_config = detector_config
        self.outputs.calibrant = calibrant
        self.outputs.mask = mask
        self.outputs.flatfield = flatfield
        self.outputs.darkcurrent = darkcurrent
        self.outputs.integration_options = integration_options

    def merged_integration_options(self) -> dict:
        """Merge integration options in this order of priority:

        - filename (lowest priority)
        - filenames[0]
        - filenames[1]
        - ...
        - integration_options (highest priority)
        """
        integration_options = dict()
        filenames = list()
        if self.inputs.filename:
            filenames.append(self.inputs.filename)
        if self.inputs.filenames:
            filenames.extend(self.inputs.filenames)
        for filename in filenames:
            integration_options.update(pyfai_utils.read_config(filename))
        if self.inputs.integration_options:
            integration_options.update(
                pyfai_utils.normalize_parameters(self.inputs.integration_options)
            )
        return integration_options



import os
import numpy
from ewoks import execute_graph
from silx.io.dictdump import dicttonx, nxtodict
import pyFAI.azimuthalIntegrator
from .xrpd_theory import Calibration, Setup, Measurement, xPattern, yPattern
from .test_calibrate import guess_fit_parameters


def calibintworkflow():
    nodes = [
        {
            "id": "calib_multidistance",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.calibrate.CalibrateMulti",
        },
        {
            "id": "calib_guess",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.calibrate.CalculateGeometry",
        },
        {
            "id": "calib_singledistance",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.calibrate.CalibrateSingle",
        },
        {
            "id": "detect_mask",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.mask.MaskDetection",
        },
        {
            "id": "subtract_background",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.background.SubtractBackground",
        },
        {
            "id": "integrate1d",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.integrate.Integrate1D",
        },
        {
            "id": "calib_integrate1d",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.integrate.Integrate1D",
        },
        {
            "id": "diagnose_multicalib",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.diagnostics.DiagnoseCalibrateMultiResults",
        },
        {
            "id": "diagnose_singlecalib",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.diagnostics.DiagnoseCalibrateSingleResults",
        },
        {
            "id": "integrate1d_plot",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.diagnostics.DiagnoseIntegrate1D",
        },
        {
            "id": "diagnose_integrate1d",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.diagnostics.DiagnoseIntegrate1D",
        },
        {
            "id": "save_ascii",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.ascii.SaveAsciiPattern1D",
        },
        {
            "id": "save_nexus",
            "task_type": "class",
            "task_identifier": "ewoksxrpd.tasks.nexus.SaveNexusPattern1D",
        },
    ]
    links = [
        {
            "source": "calib_multidistance",
            "target": "calib_guess",
            "data_mapping": [
                {"source_output": "parametrization", "target_input": "parametrization"},
                {"source_output": "parameters", "target_input": "parameters"},
            ],
        },
        {
            "source": "calib_multidistance",
            "target": "diagnose_multicalib",
            "data_mapping": [
                {"source_output": "rings", "target_input": "rings"},
                {"source_output": "parametrization", "target_input": "parametrization"},
                {"source_output": "parameters", "target_input": "parameters"},
                {"source_output": "detector", "target_input": "detector"},
                {"source_output": "detector_config", "target_input": "detector_config"},
            ],
        },
        {
            "source": "calib_guess",
            "target": "calib_singledistance",
            "data_mapping": [
                {"source_output": "geometry", "target_input": "geometry"},
                {"source_output": "energy", "target_input": "energy"},
            ],
        },
        {
            "source": "calib_singledistance",
            "target": "integrate1d",
            "data_mapping": [
                {"source_output": "geometry", "target_input": "geometry"},
                {"source_output": "energy", "target_input": "energy"},
                {"source_output": "detector", "target_input": "detector"},
                {"source_output": "detector_config", "target_input": "detector_config"},
            ],
        },
        {
            "source": "calib_singledistance",
            "target": "calib_integrate1d",
            "data_mapping": [
                {"source_output": "geometry", "target_input": "geometry"},
                {"source_output": "energy", "target_input": "energy"},
                {"source_output": "detector", "target_input": "detector"},
                {"source_output": "detector_config", "target_input": "detector_config"},
            ],
        },
        {
            "source": "calib_singledistance",
            "target": "diagnose_singlecalib",
            "data_mapping": [
                {"source_output": "geometry", "target_input": "geometry"},
                {"source_output": "energy", "target_input": "energy"},
                {"source_output": "detector", "target_input": "detector"},
                {"source_output": "detector_config", "target_input": "detector_config"},
                {"source_output": "rings", "target_input": "rings"},
            ],
        },
        {
            "source": "subtract_background",
            "target": "integrate1d",
            "data_mapping": [
                {"source_output": "image", "target_input": "image"},
                {"source_output": "monitor", "target_input": "monitor"},
            ],
        },
        {
            "source": "detect_mask",
            "target": "integrate1d",
            "data_mapping": [{"source_output": "mask", "target_input": "mask"}],
        },
        {
            "source": "calib_integrate1d",
            "target": "diagnose_integrate1d",
            "data_mapping": [
                {"source_output": "x", "target_input": "x"},
                {"source_output": "y", "target_input": "y"},
                {"source_output": "xunits", "target_input": "xunits"},
            ],
        },
        {
            "source": "calib_singledistance",
            "target": "diagnose_integrate1d",
            "data_mapping": [
                {"source_output": "energy", "target_input": "energy"},
            ],
        },
        {
            "source": "integrate1d",
            "target": "integrate1d_plot",
            "data_mapping": [
                {"source_output": "x", "target_input": "x"},
                {"source_output": "y", "target_input": "y"},
                {"source_output": "xunits", "target_input": "xunits"},
            ],
        },
        {
            "source": "integrate1d",
            "target": "save_ascii",
            "data_mapping": [
                {"source_output": "x", "target_input": "x"},
                {"source_output": "y", "target_input": "y"},
                {"source_output": "yerror", "target_input": "yerror"},
                {"source_output": "xunits", "target_input": "xunits"},
                {"source_output": "info", "target_input": "header"},
            ],
        },
        {
            "source": "integrate1d",
            "target": "save_nexus",
            "data_mapping": [
                {"source_output": "x", "target_input": "x"},
                {"source_output": "y", "target_input": "y"},
                {"source_output": "yerror", "target_input": "yerror"},
                {"source_output": "xunits", "target_input": "xunits"},
                {"source_output": "info", "target_input": "header"},
            ],
        },
    ]

    return {"graph": {"id": "calint"}, "nodes": nodes, "links": links}


def test_calint_workflow(
    generate_examples,
    tmpdir,
    imageSetup1Calibrant1: Calibration,
    setup1: Setup,
    imageSetup2Calibrant1: Calibration,
    setup2: Setup,
    aiSetup1: pyFAI.azimuthalIntegrator.AzimuthalIntegrator,
    image1Setup1SampleB: Measurement,
    image2Setup1SampleB: Measurement,
    imageSetup1SampleA: Measurement,
    xSampleA: xPattern,
    ySampleA: yPattern,
):
    if generate_examples:
        from pathlib import Path
        from shutil import rmtree

        import ewoksxrpd

        tmpdir = Path(ewoksxrpd.__path__[0]).parent / "examples"
        rmtree(tmpdir / "data", ignore_errors=True)
        rmtree(tmpdir / "results", ignore_errors=True)
        rmtree(tmpdir / "transient", ignore_errors=True)

    datadir = tmpdir / "data"
    datadir.mkdir()
    resultdir = tmpdir / "results"
    resultdir.mkdir()
    transientdir = tmpdir / "transient"
    transientdir.mkdir()

    calibrant = imageSetup1Calibrant1.calibrant
    detector = setup1.detector
    geometry0, energy0 = guess_fit_parameters(
        setup1.geometry, setup1.energy, aiSetup1, []
    )

    # Calibration data
    mcalib_images, mcalib_positions = _multidistance_calibration_data(
        datadir, imageSetup1Calibrant1, setup1, imageSetup2Calibrant1, setup2
    )
    scalib_image = mcalib_images[0]
    scalib_position = mcalib_positions[0]

    # Mask data
    tnorm = image1Setup1SampleB.monitor
    maskimage1, maskmonitor1, maskimage2, maskmonitor2 = _mask_data(
        datadir, tnorm, image1Setup1SampleB, image2Setup1SampleB, setup1
    )

    # Background+sample data
    backgroundimage, background_monitor, sampleimage, samplemonitor = _holder_data(
        datadir,
        tnorm,
        imageSetup1SampleA,
        image1Setup1SampleB,
        image2Setup1SampleB,
        setup1,
    )

    refmonitor = ySampleA.monitor
    integration_options = xSampleA.integration_options

    inputs = [
        # multi distance calibration
        {"id": "calib_multidistance", "name": "images", "value": mcalib_images},
        {"id": "calib_multidistance", "name": "positions", "value": mcalib_positions},
        {"id": "calib_multidistance", "name": "positionunits_in_meter", "value": 1e-2},
        {"id": "calib_multidistance", "name": "detector", "value": detector},
        {"id": "calib_multidistance", "name": "geometry", "value": geometry0},
        {"id": "calib_multidistance", "name": "energy", "value": energy0},
        {"id": "calib_multidistance", "name": "calibrant", "value": calibrant},
        # single distance calibration plot
        {"id": "diagnose_multicalib", "name": "images", "value": mcalib_images},
        {
            "id": "diagnose_multicalib",
            "name": "positions",
            "value": mcalib_positions,
        },
        {
            "id": "diagnose_multicalib",
            "name": "positionunits_in_meter",
            "value": 1e-2,
        },
        {"id": "diagnose_multicalib", "name": "calibrant", "value": calibrant},
        {
            "id": "diagnose_multicalib",
            "name": "filename",
            "value": str(resultdir / "diagnose_multicalib.png"),
        },
        # calibration guess
        {"id": "calib_guess", "name": "position", "value": scalib_position},
        # single distance calibration
        {"id": "calib_singledistance", "name": "image", "value": scalib_image},
        {"id": "calib_singledistance", "name": "detector", "value": detector},
        {"id": "calib_singledistance", "name": "calibrant", "value": calibrant},
        {"id": "calib_singledistance", "name": "fixed", "value": ["energy"]},
        # single distance calibration plot
        {"id": "diagnose_singlecalib", "name": "image", "value": scalib_image},
        {"id": "diagnose_singlecalib", "name": "calibrant", "value": calibrant},
        {
            "id": "diagnose_singlecalib",
            "name": "filename",
            "value": str(resultdir / "diagnose_singlecalib.png"),
        },
        # detect mask
        {"id": "detect_mask", "name": "image1", "value": maskimage1},
        {"id": "detect_mask", "name": "monitor1", "value": maskmonitor1},
        {"id": "detect_mask", "name": "image2", "value": maskimage2},
        {"id": "detect_mask", "name": "monitor2", "value": maskmonitor2},
        # subtract background
        {"id": "subtract_background", "name": "image", "value": sampleimage},
        {"id": "subtract_background", "name": "monitor", "value": samplemonitor},
        {"id": "subtract_background", "name": "background", "value": backgroundimage},
        {
            "id": "subtract_background",
            "name": "background_monitor",
            "value": background_monitor,
        },
        # integrate
        {
            "id": "integrate1d",
            "name": "integration_options",
            "value": integration_options,
        },
        {"id": "integrate1d", "name": "reference", "value": refmonitor},
        # integrate calibrant
        {"id": "calib_integrate1d", "name": "image", "value": scalib_image},
        # plot integrate calibrant
        {"id": "diagnose_integrate1d", "name": "calibrant", "value": calibrant},
        {
            "id": "diagnose_integrate1d",
            "name": "filename",
            "value": str(resultdir / "diagnose_integrate1d.png"),
        },
        # save result
        {
            "id": "save_ascii",
            "name": "filename",
            "value": str(resultdir / "result_integrate1d.dat"),
        },
        {
            "id": "save_ascii",
            "name": "metadata",
            "value": {
                "name": "S-220323-00006",
                "doi": "4590be84-3493-4bd2-91fe-4cf39cfcf71f",
            },
        },
        {
            "id": "save_nexus",
            "name": "url",
            "value": str(resultdir / "result_integrate1d.h5"),
        },
        {
            "id": "save_nexus",
            "name": "metadata",
            "value": {
                "sample": {
                    "@NX_class": "NXsample",
                    "name": "S-220323-00006",
                    "doi": "4590be84-3493-4bd2-91fe-4cf39cfcf71f",
                }
            },
        },
        {
            "id": "integrate1d_plot",
            "name": "filename",
            "value": str(resultdir / "result_integrate1d.png"),
        },
    ]

    outputs = [
        {"id": "integrate1d", "name": "x"},
        {"id": "integrate1d", "name": "y"},
        {"id": "integrate1d", "name": "xunits"},
    ]

    workflow = calibintworkflow()

    result = execute_graph(
        workflow,
        inputs=inputs,
        outputs=outputs,
        varinfo={"root_uri": str(transientdir / "result.nx"), "scheme": "nexus"},
    )

    assert result["xunits"] == xSampleA.units
    numpy.testing.assert_allclose(xSampleA.x, result["x"], rtol=1e-6)
    atol = ySampleA.y.max() * 0.01
    numpy.testing.assert_allclose(ySampleA.y, result["y"], atol=atol)

    x, y, yerror = numpy.loadtxt(str(resultdir / "result_integrate1d.dat")).T
    numpy.testing.assert_array_equal(x, result["x"])
    numpy.testing.assert_array_equal(y, result["y"])
    assert numpy.isnan(yerror).all()

    adict = nxtodict(str(resultdir / "result_integrate1d.h5"))
    diffractogram = adict["results"]["integrate"]["integrated"]
    x = diffractogram["2th"]
    y = diffractogram["intensity"]
    yerror = diffractogram["intensity_errors"]
    numpy.testing.assert_array_equal(x, result["x"])
    numpy.testing.assert_array_equal(y, result["y"])
    assert numpy.isnan(yerror).all()

    assert os.path.isfile(str(resultdir / "diagnose_multicalib.png"))
    assert os.path.isfile(str(resultdir / "diagnose_singlecalib.png"))
    assert os.path.isfile(str(resultdir / "diagnose_integrate1d.png"))
    assert os.path.isfile(str(resultdir / "result_integrate1d.png"))

    if generate_examples:
        from ewokscore import load_graph

        import matplotlib.pyplot as plt

        taskgraph = load_graph(calibintworkflow(), inputs=inputs)
        taskgraph.dump(str(tmpdir / "xrpd_workflow.json"), indent=2)
        plt.show()


def _multidistance_calibration_data(
    datadir, imageSetup1Calibrant1, setup1, imageSetup2Calibrant1, setup2
):
    mcalib_images = list()
    mcalib_positions = list()
    images = [
        (
            imageSetup1Calibrant1.image,
            setup1.geometry["dist"] * 100,
        ),
        (
            imageSetup2Calibrant1.image,
            setup2.geometry["dist"] * 100,
        ),
    ]
    data = {"@NX_class": "NXroot", "@default": "1.1"}
    filename = str(datadir / "calib.h5")
    for i, (image, detz) in enumerate(images, 1):
        data[f"{i}.1"] = {
            "@default": "plotselect",
            "instrument": {
                "@NX_class": "NXinstrument",
                "pilatus1": {
                    "@NX_class": "NXdetector",
                    "data": image,
                },
                "positioners": {"detz": detz, "detz@units": "cm"},
            },
            "title": "sct 1",
            "measurement": {">pilatus1": "../instrument/pilatus1/data"},
            "plotselect": {
                "@NX_class": "NXdata",
                "@signal": "data",
                ">data": "../instrument/pilatus1/data",
            },
        }
        mcalib_images.append(f"silx://{filename}?path=/{i}.1/measurement/pilatus1")
        mcalib_positions.append(
            f"silx://{filename}?path=/{i}.1/instrument/positioners/detz"
        )
    dicttonx(data, filename, update_mode="add")

    return mcalib_images, mcalib_positions


def _mask_data(datadir, tnorm, image1Setup1SampleB, image2Setup1SampleB, setup1):
    detz = setup1.geometry["dist"] * 100

    images = [
        (image1Setup1SampleB.image, image1Setup1SampleB.monitor),
        (image2Setup1SampleB.image, image2Setup1SampleB.monitor),
    ]
    data = {"@NX_class": "NXroot", "@default": "1.1"}
    for i, (image, I0) in enumerate(images, 1):
        data[f"{i}.1"] = {
            "@default": "plotselect",
            "instrument": {
                "@NX_class": "NXinstrument",
                "pilatus1": {
                    "@NX_class": "NXdetector",
                    "data": image,
                },
                "I0": {
                    "@NX_class": "NXdetector",
                    "data": I0,
                },
                "positioners": {"detz": detz, "detz@units": "cm"},
            },
            "title": f"sct {I0/tnorm}",
            "measurement": {
                ">pilatus1": "../instrument/pilatus1/data",
                ">I0": "../instrument/I0/data",
            },
            "plotselect": {
                "@NX_class": "NXdata",
                "@signal": "data",
                ">data": "../instrument/pilatus1/data",
            },
        }
    filename = str(datadir / "mask.h5")
    dicttonx(data, filename, update_mode="add")
    maskimage1 = f"silx://{filename}?path=/1.1/measurement/pilatus1"
    maskmonitor1 = f"silx://{filename}?path=/1.1/measurement/I0"
    maskimage2 = f"silx://{filename}?path=/2.1/measurement/pilatus1"
    maskmonitor2 = f"silx://{filename}?path=/2.1/measurement/I0"

    return maskimage1, maskmonitor1, maskimage2, maskmonitor2


def _holder_data(
    datadir, tnorm, imageSetup1SampleA, image1Setup1SampleB, image2Setup1SampleB, setup1
):
    detz = setup1.geometry["dist"] * 100
    assert imageSetup1SampleA.monitor == image1Setup1SampleB.monitor
    images = [
        (image2Setup1SampleB.image, image2Setup1SampleB.monitor),
        (
            imageSetup1SampleA.image + image1Setup1SampleB.image,
            imageSetup1SampleA.monitor,
        ),
    ]
    data = {"@NX_class": "NXroot", "@default": "2.1"}
    for i, (image, I0) in enumerate(images, 1):
        data[f"{i}.1"] = {
            "@default": "plotselect",
            "instrument": {
                "@NX_class": "NXinstrument",
                "pilatus1": {
                    "@NX_class": "NXdetector",
                    "data": image,
                },
                "I0": {
                    "@NX_class": "NXdetector",
                    "data": I0,
                },
                "positioners": {"detz": detz, "detz@units": "cm"},
            },
            "title": f"sct {I0/tnorm}",
            "measurement": {
                ">pilatus1": "../instrument/pilatus1/data",
                ">I0": "../instrument/I0/data",
            },
            "plotselect": {
                "@NX_class": "NXdata",
                "@signal": "data",
                ">data": "../instrument/pilatus1/data",
            },
        }
    filename = str(datadir / "holder.h5")
    dicttonx(data, filename, update_mode="add")
    backgroundimage = f"silx://{filename}?path=/1.1/measurement/pilatus1"
    background_monitor = f"silx://{filename}?path=/1.1/measurement/I0"
    sampleimage = f"silx://{filename}?path=/2.1/measurement/pilatus1"
    samplemonitor = f"silx://{filename}?path=/2.1/measurement/I0"

    return backgroundimage, background_monitor, sampleimage, samplemonitor



from contextlib import contextmanager

import h5py
from silx.io.dictdump import dicttonx
from ewokscore.missing_data import is_missing_data
from ewoksdata.data.nexus import select_default_plot

from .utils import data_utils, pyfai_utils
from .data_access import TaskWithDataAccess

__all__ = ["SaveNexusPattern1D", "SaveNexusIntegrated"]


class _BaseSaveNexusIntegrated(
    TaskWithDataAccess,
    input_names=["url"],
    optional_input_names=[
        "bliss_scan_url",
        "metadata",
        "nxprocess_name",
        "nxmeasurement_name",
        "nxprocess_as_default",
        "external_url",
    ],
    output_names=["saved"],
    register=False,
):
    @property
    def _process_info(self):
        raise NotImplementedError

    @property
    def _nxprocess_name(self):
        if self.inputs.nxprocess_name:
            return self.inputs.nxprocess_name
        return "integrate"

    @property
    def _nxmeasurement_name(self):
        if self.inputs.nxmeasurement_name:
            return self.inputs.nxmeasurement_name
        return "integrated"

    @contextmanager
    def _save_context(self):
        link_results_nxentry_url = self.inputs.url
        results_nxentry_url = self.get_input_value(
            "external_url", link_results_nxentry_url
        )

        with self.open_h5item(
            results_nxentry_url, mode="a", create=True
        ) as results_nxentry:
            assert isinstance(results_nxentry, h5py.Group)
            with self.open_h5item(
                link_results_nxentry_url, mode="a", create=True
            ) as link_results_nxentry:
                assert isinstance(link_results_nxentry, h5py.Group)
                results_nxprocess = pyfai_utils.create_nxprocess(
                    results_nxentry,
                    link_results_nxentry,
                    self._nxprocess_name,
                    self._process_info,
                )

            yield results_nxprocess

            # Create links
            with self.open_h5item(
                link_results_nxentry_url, mode="a", create=True
            ) as link_results_nxentry:
                bliss_scan_url = data_utils.data_from_storage(
                    self.inputs.bliss_scan_url, remove_numpy=True
                )
                if bliss_scan_url:
                    self.link_bliss_scan(link_results_nxentry, bliss_scan_url)

                results_nxdata = results_nxprocess.get("integrated")
                if results_nxdata is not None:
                    if self.get_input_value("nxprocess_as_default", True):
                        select_default_plot(results_nxdata)

                    intensity = results_nxdata.get("intensity")
                    if intensity is not None:
                        nxmeasurement = results_nxentry.require_group("measurement")
                        nxmeasurement.attrs.setdefault("NX_class", "NXcollection")
                        data_utils.create_hdf5_link(
                            nxmeasurement, self._nxmeasurement_name, intensity
                        )

                        link_nxmeasurement = link_results_nxentry.require_group(
                            "measurement"
                        )
                        link_nxmeasurement.attrs.setdefault("NX_class", "NXcollection")
                        data_utils.create_hdf5_link(
                            link_nxmeasurement, self._nxmeasurement_name, intensity
                        )
                        if self.inputs.metadata:
                            dicttonx(
                                self.inputs.metadata,
                                link_results_nxentry,
                                update_mode="add",
                                add_nx_class=True,
                            )
        self.outputs.saved = True


class SaveNexusPattern1D(
    _BaseSaveNexusIntegrated,
    input_names=["x", "y", "xunits"],
    optional_input_names=["header", "yerror"],
):
    """Save single diffractogram in HDF5/NeXus format"""

    def run(self):
        with self._save_context() as results_nxprocess:
            results_nxdata = pyfai_utils.create_integration_results_nxdata(
                results_nxprocess,
                self.inputs.y.ndim,
                self.inputs.x,
                self.inputs.xunits,
                None,
                None,
            )
            results_nxdata.attrs["signal"] = "intensity"
            results_nxdata["intensity"] = self.inputs.y
            if not self.missing_inputs.yerror:
                results_nxdata["intensity_errors"] = self.inputs.yerror

    @property
    def _process_info(self):
        return self.inputs.header


class SaveNexusIntegrated(
    _BaseSaveNexusIntegrated,
    input_names=["radial", "intensity", "radial_units"],
    optional_input_names=["info", "azimuthal", "intensity_error", "azimuthal_units"],
):
    """Save 1D or 2D integration diffraction patterns in HDF5/NeXus format"""

    def run(self):
        with self._save_context() as results_nxprocess:
            # Fallback for old workflows that do not specify azimuthal_units but do have azimuthal data
            if is_missing_data(self.inputs.azimuthal_units) and not is_missing_data(
                self.inputs.azimuthal
            ):
                azimuthal_units = "chi_deg"
            else:
                azimuthal_units = self.inputs.azimuthal_units
            results_nxdata = pyfai_utils.create_integration_results_nxdata(
                results_nxprocess,
                self.inputs.intensity.ndim,
                self.inputs.radial,
                self.inputs.radial_units,
                self.inputs.azimuthal,
                azimuthal_units,
            )
            results_nxdata.attrs["signal"] = "intensity"
            results_nxdata["intensity"] = self.inputs.intensity
            if not self.missing_inputs.intensity_error:
                results_nxdata["intensity_errors"] = self.inputs.intensity_error

    @property
    def _process_info(self):
        return self.inputs.info



from ewoksorange.tests.utils import execute_task
from ewoksxrpd.tasks.save_images import SaveImages
import numpy
import glob
import h5py

NUM_IMAGES = 5


def test_sum(tmpdir):
    image_uris = []
    with h5py.File(tmpdir / "output.h5", "w") as h5file:
        for i in range(NUM_IMAGES):
            image = h5file.create_dataset(str(i), data=numpy.random.random((100, 80)))
            image_uris.append(f'{tmpdir / "output.h5"}::{image.name}')

    inputs = {
        "output_dir": tmpdir,
        "image_uris": image_uris,
    }
    outputs = execute_task(
        SaveImages,
        inputs=inputs,
    )

    assert len(outputs["output_paths"]) == NUM_IMAGES
    assert sorted(glob.glob(str(tmpdir / "*.edf"))) == outputs["output_paths"]



from contextlib import contextmanager
from typing import Iterator, Union
import numpy
import h5py
from silx.io.url import DataUrl
from silx.io import h5py_utils
from ewokscore import TaskWithProgress
from ewoksdata.data import nexus
from ewoksdata.data import bliss
from .utils import data_utils


class TaskWithDataAccess(
    TaskWithProgress,
    optional_input_names=["retry_timeout", "retry_period"],
    register=False,
):
    def get_retry_options(self):
        retry_timeout = self.get_input_value("retry_timeout", None)
        retry_period = self.get_input_value("retry_period", None)
        return {"retry_timeout": retry_timeout, "retry_period": retry_period}

    @contextmanager
    def open_h5item(
        self, url: Union[str, DataUrl], create: bool = False, **openoptions
    ) -> Iterator[Union[h5py.Group, h5py.Dataset, numpy.ndarray]]:
        if isinstance(url, DataUrl):
            url = DataUrl(url)
        retryoptions = self.get_retry_options()
        if create:
            url = nexus.create_url(url, **retryoptions)
        with h5py_utils.open_item(
            url.file_path(), url.data_path(), **retryoptions, **openoptions
        ) as item:
            idx = url.data_slice()
            if idx is None:
                yield item
            else:
                yield item[idx]

    def get_data(self, *args, **kw):
        kw.update(self.get_retry_options())
        return bliss.get_data(*args, **kw)

    def get_image(self, *args, **kw):
        kw.update(self.get_retry_options())
        return bliss.get_image(*args, **kw)

    def iter_bliss_data(self, *args, **kw):
        kw.update(self.get_retry_options())
        yield from bliss.iter_bliss_scan_data(*args, **kw)

    def iter_bliss_data_from_memory(self, *args, **kw):
        kw.update(self.get_retry_options())
        yield from bliss.iter_bliss_scan_data_from_memory(*args, **kw)

    def link_bliss_scan(self, *args, **kw):
        kw.update(self.get_retry_options())
        return data_utils.link_bliss_scan(*args, **kw)



from typing import Dict, Mapping, Tuple
from silx.gui.plot import Plot1D

from ewoksxrpd.tasks.integrate import Integrate1D
from ewoksxrpd.tasks.utils import data_utils
from ewoksxrpd.gui.trigger_widget import OWTriggerWidget
from ewoksxrpd.gui.forms import input_parameters_integrate1d
from ewoksxrpd.gui.forms import pack_geometry
from ewoksxrpd.gui.forms import unpack_geometry
from ewoksxrpd.gui.forms import unpack_enabled_geometry

__all__ = ["OWIntegrate1D"]


class OWIntegrate1D(OWTriggerWidget, ewokstaskclass=Integrate1D):
    name = "Integrate1D"
    description = "Azimuthal integration of a 2D diffraction pattern"
    icon = "icons/widget.png"
    want_main_area = True

    def _init_forms(self):
        parameter_info = input_parameters_integrate1d(self.get_default_input_values())
        self._create_input_form(parameter_info)

    def _init_main_area(self):
        layout = self._get_main_layout()
        self._plot = Plot1D()
        layout.addWidget(self._plot)
        super()._init_main_area()

    def _refresh_non_form_output_widgets(self):
        with self._capture_errors():
            super()._refresh_non_form_output_widgets()
            self._update_plot()

    def _update_plot(self):
        self._plot.remove(kind="curve")

        outputs = self.get_task_output_values()
        x = outputs.get("x")
        y = outputs.get("y")
        xunits = outputs.get("xunits")
        if not xunits or not data_utils.is_data(x) or not data_utils.is_data(y):
            return

        inputs = self.get_task_input_values()
        reference = inputs.get("reference")
        self._plot.addCurve(x, y, xlabel=xunits, ylabel=f"Normalized to {reference}")

    def _values_from_form(
        self, values: Mapping, checked: Dict[str, bool], output: bool = False
    ) -> Mapping:
        return pack_geometry(values, checked)

    def _values_to_form(
        self, values: Mapping, output: bool = False
    ) -> Tuple[Mapping, Dict[str, bool]]:
        return unpack_geometry(values)

    def _enabled_to_form(
        self, enabled: Dict[str, bool], output: bool = False
    ) -> Dict[str, bool]:
        return unpack_enabled_geometry(enabled)



from __future__ import annotations
import logging
import warnings
from numbers import Number
from typing import Dict, List, Optional, Set, Tuple, Sequence, Union, Generator

import pyFAI.detectors
import pyFAI.calibrant
import pyFAI.geometry
import pyFAI.goniometer
import pyFAI.geometryRefinement

from .utils import xrpd_utils
from .utils import data_utils
from .data_access import TaskWithDataAccess


__all__ = ["CalibrateSingle", "CalibrateMulti", "CalculateGeometry"]


logger = logging.getLogger(__name__)


def parse_fixed(
    fixed: Optional[List[str]], parametrization: Dict[str, List[str]]
) -> Set[str]:
    if not fixed:
        return set()
    allowed = {"dist", "poni1", "poni2", "rot1", "rot2", "rot3", "energy"}
    existing = set(fixed)
    unexpected = existing - allowed
    if unexpected:
        raise ValueError(f"'fixed' has unexpected parameters {sorted(unexpected)}")
    out = set()
    for param in fixed:
        out |= set(parametrization.get(param, {param}))
    return out


class CalibrateSingle(
    TaskWithDataAccess,
    input_names=["image", "detector", "calibrant", "geometry", "energy"],
    optional_input_names=[
        "detector_config",
        "fixed",
        "max_rings",
        "robust",
        "ring_detector",
    ],
    output_names=["geometry", "energy", "detector", "detector_config", "rings", "chi2"],
):
    """Single distance and energy calibration."""

    @property
    def _max_rings(self):
        max_rings = self.inputs.max_rings
        if not max_rings:
            return None
        if max_rings < 0:
            return None
        return max_rings

    def run(self):
        detector = data_utils.data_from_storage(self.inputs.detector)
        detector_config = data_utils.data_from_storage(
            self.get_input_value("detector_config", None)
        )
        detector_object = pyFAI.detectors.detector_factory(
            detector, config=detector_config
        )
        calibrant = data_utils.data_from_storage(self.inputs.calibrant)
        calibrant_object = pyFAI.calibrant.get_calibrant(calibrant)
        geometry0 = data_utils.data_from_storage(self.inputs.geometry)
        xrpd_utils.validate_geometry(geometry0)
        wavelength0 = xrpd_utils.energy_wavelength(self.inputs.energy)

        # Find diffraction rings and fit
        geometry_refinement = self._refine(
            dict(geometry0), wavelength0, calibrant_object, detector_object
        )

        # Parse results
        results = geometry_refinement.get_config()
        geometry = {
            k: v for k, v in results.items() if k in xrpd_utils.GEOMETRY_PARAMETERS
        }
        energy = xrpd_utils.energy_wavelength(results["wavelength"])
        rings = xrpd_utils.points_to_rings(geometry_refinement.data, calibrant_object)

        self.outputs.geometry = geometry
        self.outputs.energy = energy
        self.outputs.detector = detector
        self.outputs.detector_config = detector_config
        self.outputs.rings = rings
        if geometry_refinement.data.ndim != 2:
            self.outputs.chi2 = float("nan")
        else:
            self.outputs.chi2 = geometry_refinement.chi2()

    def _refine(
        self,
        geometry0: dict,
        wavelength0: float,
        calibrant: pyFAI.calibrant.Calibrant,
        detector: pyFAI.detectors.Detector,
    ) -> pyFAI.GeometryRefinement.GeometryRefinement:
        """Find diffraction rings and fit"""
        parametrization = {"energy": ["wavelength"]}
        force_fixed = parse_fixed(self.inputs.fixed, parametrization)
        all_parameters = xrpd_utils.GEOMETRY_PARAMETERS | {"wavelength"}

        if self.inputs.robust:
            parameter_release_groups = [
                {"dist"},
                {"poni1", "poni2"},
                {"rot1", "rot2", "rot3"},
                {"wavelength"},
            ]
        else:
            parameter_release_groups = [set(all_parameters)]

        # Initial fit parameters
        geometry = geometry0
        wavelength = wavelength0

        n = len(parameter_release_groups)
        it_fixed_parameters = _iter_fixed_parameters(
            all_parameters, force_fixed, parameter_release_groups
        )
        for i, fixed_parameters in enumerate(it_fixed_parameters, 1):
            # Detect experimental rings
            cps = self._extract_control_points(
                geometry, calibrant, wavelength, detector
            )

            # Fit parameters to detected rings
            setup = pyFAI.goniometer.SingleGeometry(
                "single_user_detector",
                detector=detector,
                geometry=geometry,
                calibrant=calibrant,
                control_points=cps,
            )
            setup.set_wavelength(wavelength)

            if cps is None:
                logger.warning("No rings detected for calibration")
                return setup.geometry_refinement

            _refine_geometry(
                setup.geometry_refinement, fixed_parameters=fixed_parameters
            )

            # Update parameters with fit result
            results = setup.geometry_refinement.get_config()
            geometry = {
                k: v for k, v in results.items() if k in xrpd_utils.GEOMETRY_PARAMETERS
            }
            wavelength = results["wavelength"]

            self.progress = i / n * 100

        return setup.geometry_refinement

    def _extract_control_points(
        self,
        geometry: dict,
        calibrant: pyFAI.calibrant.Calibrant,
        wavelength: float,
        detector: pyFAI.detectors.Detector,
    ) -> pyFAI.goniometer.ControlPoints | None:
        ring_detector_name = self.get_input_value("ring_detector", None)
        if ring_detector_name is None:
            ring_detector = detector
        else:
            ring_detector = pyFAI.detectors.detector_factory(ring_detector_name)

        setup = pyFAI.goniometer.SingleGeometry(
            "single_ring_detector",
            image=self.get_image(self.inputs.image),
            detector=ring_detector,
            geometry=geometry,
            calibrant=calibrant,
        )
        setup.set_wavelength(wavelength)

        setup.extract_cp(max_rings=self._max_rings)
        if setup.geometry_refinement.data.ndim != 2:
            logger.warning("No rings detected for calibration")
            return

        return setup.control_points


class CalibrateMulti(
    TaskWithDataAccess,
    input_names=[
        "images",
        "positions",
        "detector",
        "calibrant",
        "geometry",
        "energy",
    ],
    optional_input_names=[
        "reference_position",
        "sample_position",
        "positionunits_in_meter",
        "detector_config",
        "fixed",
        "max_rings",
        "robust",
    ],
    output_names=[
        "geometry",
        "energy",
        "detector",
        "detector_config",
        "rings",
        "chi2",
        "parametrization",
        "parameters",
    ],
):
    """Single energy, multi distance calibration with 1 distance motor.

    The `images` and `positions` are the diffraction patterns and the motor positions at which those were measured.

    The `reference_position` is the detector position that corresponds to `geometry` (`positions[0]` by default).

    The `sample_position` is the position where sample and detector would theoretically coincide (0 by default).

    The output `energy` and `geometry` correspond to `sample_position`.

    The units of `positions`, `sample_position` and `reference_position` are defined by `positionunits_in_meter`
    (1e-3 by default which means the unit is in millimeter).

    The parametrization allow deriving energy+geometry from the detector position.

    .. code-block:: python

        energy, geometry = parametrization(parameters, position)
    """

    def run(self):
        if len(self.inputs.images) != len(self.inputs.positions):
            raise ValueError("number of 'images' and 'positions' must be the same")
        detector = data_utils.data_from_storage(self.inputs.detector)
        detector_config = data_utils.data_from_storage(
            self.get_input_value("detector_config", None)
        )
        detector_object = pyFAI.detectors.detector_factory(
            detector, config=detector_config
        )
        calibrant = data_utils.data_from_storage(self.inputs.calibrant)
        calibrant_object = pyFAI.calibrant.get_calibrant(calibrant)
        energy0 = self.inputs.energy
        wavelength0 = xrpd_utils.energy_wavelength(energy0)
        geometry0 = data_utils.data_from_storage(self.inputs.geometry)
        xrpd_utils.validate_geometry(geometry0)

        # Parameterize the detector geometry
        trans_function, initial_guess = self._reparametrization(geometry0, energy0)

        # Define setup
        setup = pyFAI.goniometer.GoniometerRefinement(
            initial_guess,
            pos_function=lambda x: x,
            trans_function=trans_function,
            detector=detector_object,
            wavelength=wavelength0,
        )
        for i, (image, xi) in enumerate(zip(self.inputs.images, self.inputs.positions)):
            label = f"position{i}"
            image = self.get_image(image)
            xi = self.get_data(xi)
            setup.new_geometry(
                label, image=image, metadata=xi, calibrant=calibrant_object
            )

        # Find diffraction rings and fit
        self._refine(setup)

        # Parse results
        geometry, energy = self._get_fit_results(setup)
        paramerization = self._get_parameterization(setup)
        parameters = self._get_parameters(setup)
        rings = self._get_rings(setup, calibrant_object)
        chi2 = setup.chi2()

        self.outputs.geometry = geometry
        self.outputs.energy = energy
        self.outputs.detector = detector
        self.outputs.detector_config = detector_config
        self.outputs.rings = rings
        self.outputs.chi2 = chi2
        self.outputs.parametrization = paramerization
        self.outputs.parameters = parameters

    def _reparametrization(
        self, geometry0: Dict, energy0: float
    ) -> Tuple[pyFAI.goniometer.ExtendedTransformation, List[Number]]:
        # xunits are the position units
        # dunits are the distance units of the fit parameters
        # aunits are the angular units of the fit parameters
        dunits_per_m = 100  # dunits/m
        aunits_per_rad = 10  # aunits/rad
        m_per_xunits = self._get_position_units()  # m/xunits

        trans_function = pyFAI.goniometer.ExtendedTransformation(
            param_names=[
                "dist_offset",  # dunits
                "dist_scale",  # dunits/xunits
                "poni1_offset",  # dunits
                "poni1_scale",  # dunits/xunits
                "poni2_offset",  # dunits
                "poni2_scale",  # dunits/xunits
                "arot1",  # aunits
                "arot2",  # aunits
                "arot3",  # aunits
                "energy",  # keV
            ],
            pos_names=["x"],  # xunits
            dist_expr=f"dist_scale / {dunits_per_m} * x + dist_offset / {dunits_per_m}",
            poni1_expr=f"poni1_scale / {dunits_per_m} * x + poni1_offset / {dunits_per_m}",
            poni2_expr=f"poni2_scale / {dunits_per_m} * x + poni2_offset / {dunits_per_m}",
            rot1_expr=f"arot1 / {aunits_per_rad}",
            rot2_expr=f"arot2 / {aunits_per_rad}",
            rot3_expr=f"arot3 / {aunits_per_rad}",
            wavelength_expr="hc*1e-10/energy",
        )

        xref = self._get_reference_position()  # xunits
        dist_offset = self._get_sample_position() * m_per_xunits  # m
        dist_scale = (geometry0["dist"] - dist_offset) / xref  # m/xunits

        poni1_scale = 0.0  # m/xunits
        poni1_offset = geometry0["poni1"]  # m
        poni2_scale = 0.0  # m/xunits
        poni2_offset = geometry0["poni2"]  # m

        initial_guess = [
            dist_offset * dunits_per_m,
            dist_scale * dunits_per_m,
            poni1_offset * dunits_per_m,
            poni1_scale * dunits_per_m,
            poni2_offset * dunits_per_m,
            poni2_scale * dunits_per_m,
            geometry0["rot1"] * aunits_per_rad,
            geometry0["rot2"] * aunits_per_rad,
            geometry0["rot3"] * aunits_per_rad,
            energy0,
        ]
        return trans_function, initial_guess

    @property
    def _max_rings(self):
        max_rings = self.inputs.max_rings
        if not max_rings:
            return None
        if max_rings < 0:
            return None
        return max_rings

    def _refine(self, setup: pyFAI.goniometer.GoniometerRefinement) -> None:
        """Find diffraction rings and fit"""
        parametrization = {
            "dist": ["dist_offset", "dist_scale"],
            "poni1": ["poni1_offset", "poni1_scale"],
            "poni2": ["poni2_offset", "poni2_scale"],
            "rot1": ["arot1"],
            "rot2": ["arot2"],
            "rot3": ["arot3"],
        }
        force_fixed = parse_fixed(self.inputs.fixed, parametrization)
        all_parameters = set(setup.trans_function.param_names)

        if self.inputs.robust:
            parameter_release_groups = [
                {"dist_offset", "dist_scale"},
                {"poni1_offset", "poni2_offset"},
                {"arot1", "arot2", "arot3"},
                {"poni1_scale", "poni2_scale"},
                {"energy"},
            ]
        else:
            parameter_release_groups = [set(all_parameters)]

        n = len(parameter_release_groups)
        it_fixed_parameters = _iter_fixed_parameters(
            all_parameters, force_fixed, parameter_release_groups
        )
        for i, fixed_parameters in enumerate(it_fixed_parameters, 1):
            # Detect experimental rings
            for setup_single_dist in setup.single_geometries.values():
                setup_single_dist.extract_cp(max_rings=self._max_rings)
                if setup_single_dist.geometry_refinement.data.ndim != 2:
                    logger.warning("No rings detected for calibration")
                    return

            # Fit parameters to detected rings
            _refine_geometry(setup, fixed_parameters=fixed_parameters)

            self.progress = i / n * 100

    def _get_rings(
        self,
        setup: pyFAI.goniometer.GoniometerRefinement,
        calibrant: pyFAI.calibrant.Calibrant,
    ) -> Dict[str, dict]:
        rings = dict()
        for i, setup_single_dist in enumerate(setup.single_geometries.values()):
            points = setup_single_dist.geometry_refinement.data
            rings[str(i)] = xrpd_utils.points_to_rings(points, calibrant)
        return rings

    def _get_fit_results(
        self, setup: pyFAI.goniometer.GoniometerRefinement
    ) -> Tuple[dict, Number]:
        parameters = setup.param
        xref = self._get_reference_position()
        results = setup.trans_function(parameters, xref)
        geometry = {k: getattr(results, k) for k in xrpd_utils.GEOMETRY_PARAMETERS}
        energy = xrpd_utils.energy_wavelength(results.wavelength)
        return geometry, energy

    def _get_parameterization(
        self, setup: pyFAI.goniometer.GoniometerRefinement
    ) -> dict:
        parametrization = dict(setup.trans_function.to_dict())
        parametrization.pop("constants")
        parametrization.pop("content")
        return parametrization

    def _get_parameters(self, setup: pyFAI.goniometer.GoniometerRefinement) -> dict:
        parameters = setup.param
        names = setup.trans_function.param_names
        return dict(zip(names, parameters))

    def _get_reference_position(self):
        """first image position by default (xunits)"""
        if self.missing_inputs.reference_position:
            return self.get_data(self.inputs.positions[0])
        else:
            return self.get_data(self.inputs.reference_position)

    def _get_sample_position(self):
        """0 by default (xunits)"""
        if self.missing_inputs.sample_position:
            return 0
        else:
            return self.get_data(self.inputs.sample_position)

    def _get_position_units(self):
        """millimeter by default"""
        if self.inputs.positionunits_in_meter:
            return self.inputs.positionunits_in_meter  # m/xunits
        else:
            return 1e-3  # m/xunits


def calculate_geometry(
    parametrization: dict, parameters: dict, position: Number
) -> Tuple[dict, Number]:
    trans_function = pyFAI.goniometer.ExtendedTransformation(**parametrization)
    parameters = [parameters[name] for name in trans_function.param_names]
    trans_function(parameters, position)
    results = trans_function(parameters, position)
    geometry = {k: getattr(results, k) for k in xrpd_utils.GEOMETRY_PARAMETERS}
    energy = xrpd_utils.energy_wavelength(results.wavelength)
    return geometry, energy


class CalculateGeometry(
    TaskWithDataAccess,
    input_names=["parametrization", "parameters", "position"],
    output_names=["geometry", "energy"],
):
    """Calculate energy and geometry from pyFAI parametrization"""

    def run(self):
        parametrization = data_utils.data_from_storage(self.inputs.parametrization)
        geometry, energy = calculate_geometry(
            parametrization,
            self.inputs.parameters,
            self.get_data(self.inputs.position),
        )
        self.outputs.geometry = geometry
        self.outputs.energy = energy


def _iter_fixed_parameters(
    all_parameters: Set[str],
    force_fixed: Set[str],
    parameter_release_groups: Sequence[Set[str]],
) -> Generator[Optional[Set[str]], None, None]:
    """Start with all parameters fixed. Release parameters in groups and yield the fixed parameters after each iteration.

    :param all_parameters: all parameters
    :param force_fixed: parameters that are always fixed
    :param parameter_release_groups: groups of parameters to release in each iteration
    :yield: fixed parameters for each iteration
    """
    fixed_parameters = all_parameters
    for release_group in parameter_release_groups:
        release = release_group - force_fixed
        if release:
            fixed_parameters -= release
            yield fixed_parameters
        else:
            yield None


def _refine_geometry(
    geometry_refinement: Union[
        pyFAI.goniometer.GoniometerRefinement,
        pyFAI.geometryRefinement.GeometryRefinement,
    ],
    fixed_parameters: Optional[Sequence] = None,
) -> bool:
    """
    Refines the given geometry within the default parameter bounds.

    :param geometry_refinement: A pyFAI object that parameterizes a single or multi-distances geometry.
    :param fixed_parameters: Parameters that are not allowed to change during refinement.
    :return: True if refinement is successful within the default parameter bounds, False otherwise.
    """
    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.filterwarnings(
            "always",
            message="Values in x were outside bounds during a minimize step, clipping to bounds",
            category=RuntimeWarning,
        )
        geometry_refinement.refine3(fix=fixed_parameters)  # method="slsqp"
        if not caught_warnings:
            return True

    return False



import os
from typing import Optional

from AnyQt import QtWidgets
from AnyQt import QtGui
from AnyQt.QtCore import Qt, QPoint


class ImageViewer(QtWidgets.QGraphicsView):
    """Image viewer with panning and zooming"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self._scene = QtWidgets.QGraphicsScene(self)
        self.setScene(self._scene)

        self.setTransformationAnchor(self.AnchorUnderMouse)
        self.setResizeAnchor(self.AnchorUnderMouse)
        self.setDragMode(self.ScrollHandDrag)
        self._last_mouse_pos: Optional[QPoint] = None

    def load_image(self, image_path: Optional[str]) -> None:
        """Load an image from file"""
        if image_path and os.path.isfile(image_path):
            pixmap = QtGui.QPixmap(image_path)
            self._scene.clear()
            self._scene.addPixmap(pixmap)
        else:
            self._scene.clear()

    def wheelEvent(self, event) -> None:
        """Zoom in or out with mouse wheel"""
        # Center on the mouse position
        view_pos = event.pos()
        scene_pos_before_scale = self.mapToScene(view_pos)
        self.centerOn(scene_pos_before_scale)

        # Scale
        factor = 1.1
        if event.angleDelta().y() < 0:
            factor = 0.9
        self.scale(factor, factor)

        # Shift in position due to scaling
        scene_pos_after_scale = self.mapToScene(view_pos)
        shift_by_scaling = scene_pos_after_scale - scene_pos_before_scale

        # Undo the shift due to scaling
        center_after_scale = self.mapToScene(self.viewport().rect().center())
        self.centerOn(center_after_scale - shift_by_scaling)

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        """Store initial mouse position for panning"""
        if event.button() == Qt.LeftButton:
            self._last_mouse_pos = event.pos()

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        """Clear last mouse position on release"""
        if event.button() == Qt.LeftButton:
            self._last_mouse_pos = None

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:
        """Pan the view with left mouse button pressed"""
        if event.buttons() != Qt.LeftButton or self._last_mouse_pos is None:
            return
        delta = event.pos() - self._last_mouse_pos
        self._last_mouse_pos = event.pos()

        hscrollbar = self.horizontalScrollBar()
        hpos = hscrollbar.value()
        hscrollbar.setValue(hpos - delta.x())

        vscrollbar = self.verticalScrollBar()
        vpos = vscrollbar.value()
        vscrollbar.setValue(vpos - delta.y())
